import numpy as np
import networkx as nx
from sklearn.base import BaseEstimator, TransformerMixin


def lovain_modularity(G):
    import networkx.algorithms.community as nx_comm

    return nx_comm.modularity(G, nx_comm.louvain_communities(G))

class ThresholdNetworkMeasures(BaseEstimator, TransformerMixin):
    '''This class is designed for thresholding and then extracting network
    measures from an input correlation matrix.

    Parameters
    -----------
    threshold : float, optional
        A floating point threshold between 0 and 1.
        This represents the threshold at which a connection
        in the passed data needs to be in order for it to
        be set as an edge. The type of threshold_method
        also changes how this threshold behaves.

        If 'density', then this value represents the
        percent of edges to keep, out of all possible edges.

        ::

            default = .2

    threshold_type : {'abs', 'pos', 'neg'}
        The type of thresholding, e.g., should the threshold
        be applied to:

        - 'abs'
            The absolute value of the connections

        - 'pos'
            Only consider edges as passed the threshold if >= self.threshold

        - 'neg'
            Only consider edges as passed the if <= self.threshold

        ::

            default = 'abs'

    threshold_method : {'value', 'density'}, optional
        The method for thresholding. The two
        defined options are either to define an edge
        strictly by value, e.g., if threshold_type is 'abs',
        and threshold is .2, then any connection greater than or
        equal to .2 will be set as an edge.

        Alternatively, you may specify that the threshold be
        treated as a density. What this means is that if the threshold
        is set to for example .2, then the top 20% of edges by weight
        will be set as an edge, regardless of the actual value of the edge.

        The passed percentage will be considered
        out of all the possible edges. This will be used to
        select a threshold value, rounding up if needed, then
        all edges above or equal to the threshold will be kept
        (positive or abs case) or in neg case, all edges less than or equal.

        ::

            default = 'value'

    to_compute : valid_measure or list of, optional
        | Either a single str representing a network
            measure to compute, or a list of valid
            measures. You may also pass any custom function
            which accepts one argument G, and returns
            a value.

        | The following global measures are currently implemented
            as options:
        |


        - 'avg_cluster':
            :func:`networkx.algorithms.cluster.average_clustering`
        - 'assortativity':
            :func:`networkx.algorithms.assortativity.degree_assortativity_coefficient`
        - 'global_eff':
            :func:`networkx.algorithms.efficiency_measures.global_efficiency`
        - 'local_eff':
            :func:`networkx.algorithms.efficiency_measures.local_efficiency`
        - 'sigma':
            :func:`networkx.algorithms.smallworld.sigma`
        - 'omega`:
            :func:`networkx.algorithms.smallworld.omega`
        - 'transitivity':
            :func:`networkx.algorithms.cluster.transitivity`

        |
        | You may also select from one of the following
            averages of local measures:
        |

        - 'avg_eigenvector_centrality':
            :func:`networkx.algorithms.centrality.eigenvector_centrality_numpy`
        - 'avg_closeness_centrality':
            :func:`networkx.algorithms.centrality.closeness_centrality`
        - 'avg_degree':
            Average graph degree.
        - 'avg_triangles':
            :func:`networkx.algorithms.cluster.triangles`
        - 'avg_pagerank':
            :func:`networkx.algorithms.link_analysis.pagerank_alg.pagerank`
        - 'avg_betweenness_centrality':
            :func:`networkx.algorithms.centrality.betweenness_centrality`
        - 'avg_information_centrality':
            :func:`networkx.algorithms.centrality.information_centrality`
        - 'avg_shortest_path_length':
            :func:`networkx.algorithms.shortest_paths.generic.average_shortest_path_length`


        ::

            default = 'avg_degree'

    ensure : {None, 'edge', 'is_conn', float}, optional
        When constructing the thresholded network,
        you may optionally specify what trait of the network to
        ensure according to the passed threshold and threshold type.
        By default, this value will be set to 'edge', which means,
        if the passed settings do not ensure atleast 1 edge, the
        threshold will be incremented by the ensure_incr, default = .01,
        until there is atleast one edge.

        Alternatively, you may pass 'is_conn' to specify that
        the network should be connected, that is, every node can be reached by
        every other. Or None, to do
        no checks (which may result in some kind of error).

        Lastly, you may pass a float value between 0 and 1, which
        specifies that the criteria check the size of the largest connected
        component relative to the total number of nodes. For example, passing
        1 is the same as passing 'is_conn', where the largest connected component is
        the same size as the network, and a value of .5 would be that atleast
        half the nodes are connected to each other.

        ::

            default = 'edge'

    ensure_incr : float, optional
        The amount in which to increment each time when checking
        for an ensure criteria. Note if ensure is None, this
        parameter is not used. Also note that this value in the case of threshold_method by value,
        is repeatedly decremented from the original passed
        threshold, unless the threshold type is 'neg', in that case it will be incremented.

        In the case of threshold_type is density, this value will always be incremented,
        as only positive thresholds make sense for density.

        ::

            default = .01

    '''

    def __init__(self, threshold=.2,
                 threshold_type='abs',
                 threshold_method='value',
                 to_compute='avg_degree',
                 ensure='edge',
                 ensure_incr=.01):

        self.threshold = threshold
        self.threshold_type = threshold_type
        self.threshold_method = threshold_method
        self.to_compute = to_compute
        self.ensure = ensure
        self.ensure_incr = ensure_incr

        self._param_checks()

    def _param_checks(self):

        # Input validation
        if self.threshold_method not in ['value', 'density']:
            raise RuntimeError('threshold method must be value or density!')

        # Bad Combo
        if self.threshold_method == 'density' and self.threshold <= 0:
            raise RuntimeError('When using density, the passed threshold must be greater than 0.')

        # Bad params
        if isinstance(self.ensure, float):
            if self.ensure > 1 or self.ensure < 0:
                raise RuntimeError('ensure must be between 0 and 1')

    @property
    def feat_names_(self):
        '''The list of feature names returned
        by this objects transform function. This property
        is special in that it can interact with :class:`BPt.Loader`,
        passing along feature name information.
        '''
        if hasattr(self, '_feat_names'):
            return self._feat_names
        return None

    @feat_names_.setter
    def feat_names_(self, feat_names):
        self._feat_names = feat_names

    def fit(self, X=None, y=None):
        '''X is a 2d correlation matrix'''

        # Check here also since this estimator can be
        # used nested in pipeline w/ param searches or
        # outside of ML
        self._param_checks()
        
        # Store passed as initiaL
        self._threshold = self.threshold

        if isinstance(self.to_compute, str):
            self.to_compute = [self.to_compute]

        # The dictionary of valid options
        self._func_dict = {
            'avg_cluster': (nx.average_clustering, False),
            'assortativity': (nx.degree_assortativity_coefficient, False),
            'global_eff': (nx.global_efficiency, False),
            'local_eff': (nx.local_efficiency, False),
            'sigma': (nx.sigma, False),
            'omega': (nx.omega, False),
            'transitivity': (nx.transitivity, False),
            'lovain_modularity' : (lovain_modularity, False),
            'avg_eigenvector_centrality': (nx.eigenvector_centrality_numpy, True),
            'avg_closeness_centrality': (nx.closeness_centrality, True),
            'avg_degree': (self._avg_degree, False),
            'avg_triangles': (nx.triangles, True),
            'avg_pagerank': (nx.pagerank, True),
            'avg_betweenness_centrality': (nx.betweenness_centrality, True),
            'avg_information_centrality': (nx.information_centrality, True),
            'avg_shortest_path_length': (nx.average_shortest_path_length, False),
            'eigenvector_centrality': (nx.eigenvector_centrality_numpy, False),
        }

        return self

    def fit_transform(self, X, y=None):
        '''Fit, then transform a passed 2D numpy correlation matrix.

        Parameters
        ----------
        X : numpy array
            A 2D numpy array representing an input correlation
            matrix.

        Returns
        ---------
        X_trans : numpy array
            Returns a flat array of length number of
            measures in parameter to_compute, representing
            the calculated network statistics.
        '''

        return self.fit(X, y).transform(X)

    def _apply_threshold(self, X):

        # If not using fit
        if not hasattr(self, '_threshold'):
            self._threshold = self.threshold

        # Process threshold type on copy of X
        X_t = X.copy()
        if self.threshold_type == 'abs':
            X_t = np.abs(X_t)

        # If Value
        if self.threshold_method == 'value':
            if self.threshold_type == 'neg':
                return np.where(X_t <= self._threshold, 1, 0)
            return np.where(X_t >= self._threshold, 1, 0)

        elif self.threshold_method == 'density':

            # Rounded up
            top_n = round(X_t.shape[0] * X_t.shape[1] * self._threshold) - 1

            # If less than 0, set to 0
            if top_n < 0:
                top_n = 0

            # If neg, sort differently
            reverse = False if self.threshold_type == 'neg' else True
            thresh = sorted(X_t.flatten(), reverse=reverse)[top_n]

            # Neg and pos case
            if self.threshold_type == 'neg':
                return np.where(X_t <= thresh, 1, 0)
            return np.where(X_t >= thresh, 1, 0)

        raise RuntimeError(str(self.threshold_method) + ' not a valid.')

    def apply_threshold(self, X, to_nx=True):

        # Apply threshold to X
        X = self._apply_threshold(X)

        # Return either np array or nx graph
        if to_nx:
            return nx.from_numpy_array(X)
        return X

    def _incr_thresh(self):

        # Density case
        if self.threshold_method == 'density':
            self._threshold += self.ensure_incr
            return

        # Value case
        if self.threshold_type == 'neg':
            self._threshold += self.ensure_incr
        else:
            self._threshold -= self.ensure_incr

    def _lc_ratio(self, X):

        G = self.apply_threshold(X, to_nx=True)
        largest_cc = len(max(nx.connected_components(G), key=len))

        return largest_cc / G.number_of_nodes()

    def _is_conn(self, X):

        G = self.apply_threshold(X, to_nx=True)
        return nx.is_connected(G)

    def threshold_check(self, X):

        # TODO change to input check func
        # that can accept more format, vs just squeeze
        X = np.squeeze(X)

        self._threshold = self.threshold
        
        # No ensure passed
        if self.ensure is None:
            return self._threshold

        # Connected component size
        elif isinstance(self.ensure, float):

            # While the ratio is less than desired
            while self._lc_ratio(X) <= self.ensure:
                self._incr_thresh()

        # Keep incrementing until connected
        elif self.ensure == 'is_conn':
            while not self._is_conn(X):
                self._incr_thresh()

        # Make sure atleast one edge
        elif self.ensure == 'edge':
            while np.sum(self.apply_threshold(X, to_nx=False)) == 0:
                self._incr_thresh()

        else:
            raise RuntimeError('Invalid option passed for ensure.')

        
        # Return the maybe changed threshold
        return self._threshold

    def transform(self, X):
        '''Transform a passed 2D numpy correlation matrix.

        Parameters
        ----------
        X : numpy array
            A 2D numpy array representing an input correlation
            matrix.

        Returns
        ---------
        X_trans : numpy array
            Returns a flat array of length number of
            measures in parameter to_compute, representing
            the calculated network statistics.
        '''

        # TODO change to input check func
        # that can accept more format, vs just squeeze
        X = np.squeeze(X)

        # Perform threshold check based on the different settings
        self.threshold_check(X)

        # Apply threshold and cast to network
        G = self.apply_threshold(X, to_nx=True)

        # Compute metrics + keep track of feature names
        X_trans = []
        self.feat_names_ = []

        for compute in self.to_compute:
            
            # If custom function
            if compute not in self._func_dict:
                computed = compute(G)
                feat_name = compute.__name__

            # If pre-defined
            else:
                computed = self._compute(G, self._func_dict[compute])
                feat_name = compute

            # Keep track in X_trans
            X_trans += [computed]

            # If local measure
            if isinstance(computed, list):
                self.feat_names_ += [f'{feat_name}_{i}' for i in range(len(computed))]
            else:
                self.feat_names_ += [feat_name]
        
        # Stack feat name
        self.feat_names_ = np.hstack(self.feat_names_)

        # Return stacked metrics
        return np.hstack(X_trans)

    def _avg_degree(self, G):
        avg_degree = np.mean([i[1] for i in nx.degree(G)])
        return avg_degree

    def _compute(self, G, func_avg):

        # Unpack function
        func, avg = func_avg[0], func_avg[1]

        # Run
        computed = func(G)

        # If dict, then we need to either
        # average, or return as per node
        if isinstance(computed, dict):
            
            # Get as just list of values
            values = list(computed.values())
            
            # Average case
            if avg:
                return np.mean(values)
                
            return values
        
        # If single number, just return
        return computed


import numpy as np
from numpy.random import RandomState
from sklearn.base import BaseEstimator


class RandomParcellation(BaseEstimator):
    '''This object can be used to generate a random surface parcellation.

    Parameters
    -----------
    geo : list of array-like
        geo, or geometry, should be passed as a list of arrays
        (or sets, or lists), where
        each index of the outer list is associated with that vertex index.
        Each list should contain all of that vertex's valid neighbors.
        So for example, if the vertex at index 0 has neighbors 5, 8, 9, 10,
        then the first element of passed geo should be [5,8,9,10].

    n_parcels : int
        The number of random parcels to generate.
        Will generate as 1-n_parcels, with 0 reserved for
        non-parcels.

    medial_wall_inds : array-like of int or None, optional
        By passing an array-like containing valid integers.
        between 0 and the size of the parcel, this parameter
        can specify optionally index where parcels should not be
        generated.

        This parameter should be left as None if
        passing medial_wall_mask, as only one will be used.

        ::

            default = None

    medial_wall_mask : array-like of bool or None, optional
        Simmilar to medial_wall_inds, you may also specify
        the medial wall, or spots not to fill, via this
        parameter by passing a boolean mask, where 1 or True
        signifies that vertex is part of the medial wall to exclude.

        This parameter should be left as None if
        passing medial_wall_inds, as only one will be used.

        ::

            default = None

    random_state : int, None, or RandomState, optional
        A random seed to be passed to create a numpy RandomState,
        as either a numpy random state, an integer seed, or None
        to generate a random RandomState.

        ::

            default = None
    '''

    def __init__(self, geo, n_parcels, medial_wall_inds=None,
                 medial_wall_mask=None, random_state=None):

        # Set passed params
        self.geo = geo
        self.n_parcels = n_parcels
        self.medial_wall_inds = medial_wall_inds
        self.medial_wall_mask = medial_wall_mask
        self.random_state = random_state
        self.mask = None

    def get_parc(self, copy=True):
        '''Get the underlying parcellation, if avaliable.

        Parameters
        -----------
        copy : bool, optional
            If True, return a copy of the
            created parcellation.

            ::

                default = True

        Returns
        ---------
        parcellation : numpy.array
            Returns the parcellations as
            a numpy array of integers, where 0 indicates
            background, and parcel labels are set ordinarily as
            numbers 1 - number of parcels.
        '''

        if self.mask is None:
            self._generate_parc_from_params()

        if copy:
            return self.mask.copy()
        else:
            return self.mask

    def _generate_parc_from_params(self):

        # Proc by input args
        self._proc_medial_wall()
        self._proc_geo()
        self._proc_random_state()

        # Set up mask, done and flags
        self.sz = len(self._geo)
        self.reset()

        # Init
        self._init_parcels()

        # Then generate
        self.generate_parcels()

    def _proc_geo(self):
        '''Convert geo to list of np arrays. Also set so only
        consider neighbors as non medial wall, if any medial wall passed.'''
        self._geo = []
        for g in self.geo:
            valid_n = list(set(g) - self.m_wall)
            self._geo.append(np.array(valid_n))

    def _proc_medial_wall(self):

        # Proc medial wall inds
        if self.medial_wall_inds is not None:
            self.m_wall = set(list(self.medial_wall_inds))
        elif self.medial_wall_mask is not None:
            self.m_wall = set(list(np.where(self.medial_wall_mask)[0]))
        else:
            self.m_wall = set()

    def _proc_random_state(self):

        if self.random_state is None:
            self.r_state = RandomState()
        elif isinstance(self.random_state, int):
            self.r_state = RandomState(seed=self.random_state)
        else:
            self.r_state = self.random_state

    def reset(self):
        '''Reset this object, clearing the created parcellations.'''

        self.mask = np.zeros(self.sz, dtype='int16')
        self.done = self.m_wall.copy()
        self.ready, self.generated = False, False

    def _init_parcels(self):
        '''Initialize this object, getting it ready
        to create a new random parcellation..'''

        # Generate the starting locs
        valid = np.setdiff1d(np.arange(self.sz), np.array(list(self.done)))
        self.start_locs = self.r_state.choice(valid, size=self.n_parcels,
                                              replace=False)

        # Set random probs. that each loc is chosen
        self.probs = self.r_state.random(size=self.n_parcels)

    def _setup(self):
        '''This method sets the labels, initializes finished
        and sets the ready flag to True.'''

        # Generate corresponding labels w/ each loc
        self.labels = np.arange(1, self.n_parcels+1, dtype='int16')

        # Mask where if == 1, then that parcel is done
        self.finished = np.zeros(self.n_parcels, dtype='bool_')

        # Drop the first points
        self.mask[self.start_locs] = self.labels

        # Set ready flag to True
        self.ready = True

    def _choice(self):
        '''Select a valid label based on probs.'''

        msk = self.finished == 0
        probs = self.probs[msk] / np.sum(self.probs[msk])
        label = self.r_state.choice(self.labels[msk], p=probs)

        return label

    def _get_valid_neighbors(self, loc):

        ns = self._geo[loc]
        valid_ns = ns[self.mask[ns] == 0]

        return valid_ns

    def generate_parcels(self):
        '''Generate the random parcellation.'''

        if self.ready is False:
            self._setup()

        # Keep looping until every spot is filled
        while (self.finished == 0).any():
            self._add_spot()

        # Set generated flag when done
        self.generated = True

    def _add_spot(self):

        # Select which parcel to add to
        label = self._choice()

        # Determine valid starting locations anywhere in existing parcel
        current = np.where(self.mask == label)[0]
        valid = set(current) - self.done

        self._proc_spot(valid, label)

    def _proc_spot(self, valid, label):

        # If no valid choices, then set this parcel to finished
        if len(valid) == 0:
            self.finished[label-1] = 1
            return

        # Select randomly from the valid starting locs
        loc = self.r_state.choice(tuple(valid))

        # Select a valid + open neighbor
        valid_ns = self._get_valid_neighbors(loc)

        if len(valid_ns) > 0:

            # Select a valid choice, and add it w/ the right label
            choice = self.r_state.choice(valid_ns)
            self.mask[choice] = label

            # If this was the only choice, mark start loc as done
            if len(valid_ns) == 1:
                self.done.add(loc)

        # If there are no valid choices, mark as done
        else:
            self.done.add(loc)

            valid.remove(loc)
            self._proc_spot(valid, label)



import os

from .. import data_dr
from .funcs import load
from ..misc.text import clean_key
from ..misc.print import _get_print
from .space import get_space_options

# Order determined by chance of false positive and frequency
LH_KEYS = ['lh.', 'hemi-L_', 'left', 'hemi-L', 'lh_', 'lh-', 'lh',  'L_', 'L', 'l']
RH_KEYS = ['rh.', 'hemi-R_', 'right', 'hemi-R', 'rh_', 'rh-', 'rh', 'R_', 'R', 'r']


def get_surf_loc(space, hemi, key):
    '''Find the saved surface file based
    on space, hemi and key, with tolerance
    to different naming schemes.

    Parameters
    ----------
    space : str
        The space of the data, where space refers to a valid surface
        space, e.g., 'fsaverage'.
    
    hemi : str
        The hemisphere to find, can pass as 'lh' / 'rh' or
        some alternate formats, e.g., 'left' / 'right'.

    key : str or list of str
        The identifying key of the surface to load. If passing
        a list of keys, will try to find the correct surface in the
        order of the passed list, e.g., you should pass the most
        specific option first, then potentially more general options.

    Returns
    -------
    path : str or None
        The path of the saved requested surface file, or
        None if not found.
    '''

    # If passed list of keys, try each
    # in order, until loc is found
    if isinstance(key, list):
        for k in key:
            loc = get_surf_loc(space, hemi, key=k)
            if loc is not None:
                return loc

    # Try checking each sub
    for sub in ['surf', 'label']:
        loc = _get_surf_loc(space, hemi, key, sub)
        if loc is not None:
            return loc

    # If not found, return None for not found
    return None

def _get_surf_loc(space, hemi, key, sub):

    # Get base directory in data dr
    dr = os.path.join(data_dr, space, sub)

    # If no dr, return right away
    if not os.path.exists(dr):
        return None

    # Get correct hemis to searchs
    if hemi in LH_KEYS:
        hemi_keys = LH_KEYS.copy()
    elif hemi in RH_KEYS:
        hemi_keys = RH_KEYS.copy()

    # Get list of file in this directory
    files = os.listdir(dr)

    # If both a hemi key and the key of interest
    # return that full location
    match_files = []
    for hemi_key in hemi_keys:
        for file in files:
            if key in file and hemi_key in file:
                match_files.append(file)

    # None
    if len(match_files) == 0:
        return None

    # If more than one valid file found, prefer the
    # one with the shortest length, as most likely
    # to match exactly the requested value, if just 1
    # will select that one
    match_file = sorted(match_files, key=len)[0]
    
    # Return full path
    return os.path.join(dr, match_file)

def _load_medial_wall(space, hemi):

    # Get location
    loc = get_surf_loc(space=space, hemi=hemi, key=['medial_mask', 'medialwall'])

    # Error if not found
    if loc is None:
        raise RuntimeError(f'No medial wall information found for space={space}, hemi={hemi}')

    # Return loaded as bool
    return load(loc, dtype='bool')

def auto_load_rois(data, space=None, hemi=None, verbose=None, _print=None):

    from ..plotting.ref import SurfRef, VolRef

    # Get _print object
    _print = _get_print(verbose=verbose, _print=_print)

    # Standerdize hemi if not None
    if hemi is not None:
        if hemi in LH_KEYS:
            hemi = 'lh'
        elif hemi in RH_KEYS:
            hemi = 'rh'
    
    # Get most likely parcel
    parc = _get_most_likely_parcel(data)
    _print(f'Detected most likely parcel from passed ROIs: {parc}', level=1)

    # Grab space options, starting w/ default or user passed
    if space is None:
        space = 'default'
    space_options = [space] + get_space_options()

    # Try each
    for space in space_options:
        
        try:

            # Run to check to see if works, w/ errors / warnings muted
            SurfRef(space=space, parc=parc, verbose=-1).get_hemis_plot_vals(data)

            # Not the most eff. but now we re-run again, but passing the current verbose
            # state, now that we know this space works e.g., we want to avoid
            # printing verbose for all the ones that fail.
            # Note: Passing print will override default verbose.

            sr = SurfRef(space=space, parc=parc, _print=_print)
            vals = sr.get_hemis_plot_vals(data)

            # Apply hemi if not None
            if hemi is not None:
                vals = hemi[vals]

            _print(f'Extracted plot values for space: {space}', level=1)

            return vals, space
        except (RuntimeError, FileNotFoundError):
            pass

        try:

            # See SurfRef logic / comments for this
            VolRef(space=space, parc=parc, verbose=-1).get_plot_vals(data)

            # Now actually get
            ref = VolRef(space=space, parc=parc, _print=_print)
            plot_vals = ref.get_plot_vals(data)
            
            _print(f'Extracted plot values for space: {space}', level=1)

            return plot_vals, space
        except (RuntimeError, FileNotFoundError):
            pass

    raise RuntimeError(f'Could not find matching space for detected parc: {parc}')
        
def _get_most_likely_parcel(data):

    # Get some helpful tools from ref
    from ..plotting.ref import _load_mapping, _data_to_dict

    # Load all mappings
    mappings_dr = os.path.join(data_dr, 'mappings')
    files = [f for f in os.listdir(mappings_dr) if '.mapping.txt' in f]
    mappings = {f.replace('.mapping.txt', ''): _load_mapping(os.path.join(mappings_dr, f)) for f in files}

    # Get clean name versions of passed data
    feat_names = list(_data_to_dict(data).keys())
    clean_names = [clean_key(name) for name in feat_names]

    def get_basic_overlap(parcel, mappings):
        
        # Get set of clean refs
        clean_refs = set([clean_key(k) for k in mappings[parcel].keys()] +\
                         [clean_key(v) for v in mappings[parcel].values()])
        clean_refs = list(clean_refs)
        
        # Do a simple check for each parcel to try and
        # figure which one has the most
        n_found = 0
        
        # Go through each of the actual passed ROI names
        for name in clean_names:
            
            # For each one, cycle through all of the
            # possible reference keys until either one is found
            # incr and stop, or none are found
            for ref in clean_refs:
                if ref in name:
                    n_found += 1
                    continue
                    
        return n_found
    
    cnts = {parcel: get_basic_overlap(parcel, mappings) for parcel in mappings}
    mx = max(cnts.items(), key=lambda x:x[1])
    if mx[1] == 0:
        raise RuntimeError('Could not find reference parcellation to match ROI names to.')

    return mx[0]





import nibabel as nib
from ..misc.print import _get_print
import numpy as np
import pickle as pkl
import os
from nibabel.cifti2 import Cifti2BrainModel
from .. import data_dr
from ..loading import load
from .parc import clean_parcel_labels


def _remove_medial_wall(fill_cifti, parcel, index_map, _print=None):
    '''Remove the medial wall from a parcel / surface. For
    now assumes that the passed index map is for a cifti file with medial
    wall included.'''

    # Keep track of current position
    pos_offset = 0

    for index in index_map:
        if isinstance(index, Cifti2BrainModel):
            if index.surface_number_of_vertices is not None:
                _print('Remove medial wall brain structure =', index.brain_structure, level=1)

                # Get index of this hemisphere within the passed parcel
                inds = pos_offset + np.array(index.vertex_indices._indices)
                _print(f'inds of len {len(inds)} with pos offset = {pos_offset}', level=2)
                
                # Fill in between index_offset and index.index_offset+index.index_count these
                # vertex
                fill_cifti[index.index_offset: index.index_offset+index.index_count] = parcel[inds]
                _print(f'Filled in cifti file index [{index.index_offset}: {index.index_offset+index.index_count}]', level=2)

                # Increment to pos_offset
                pos_offset += index.surface_number_of_vertices

    return fill_cifti

def _add_subcortical(fill_cifti, index_map, _print=None):
    '''Assumes base parcel already added'''

    # Start at 1 plus last
    cnt = np.max(np.unique(fill_cifti)) + 1
    _print(f'Adding subcort ROIs starting at index {cnt}', level=1)

    for index in index_map:
        if isinstance(index, Cifti2BrainModel):
            
            # If subcort volume
            if index.surface_number_of_vertices is None:
                _print('Adding brain_structure =', index.brain_structure, level=1)
                start = index.index_offset 
                end = start + index.index_count
                fill_cifti[start:end] = cnt
                cnt += 1

    return fill_cifti

def _static_parc_to_cifti(parcel, index_map, add_sub=True, _print=None):

    # Init empty cifti with zeros
    if add_sub:
        fill_cifti = np.zeros(91282)
    else:
        fill_cifti = np.zeros(59412)

    _print(f'Init fill cifti with shape: {fill_cifti.shape}', level=1)
    _print(f'Init parcel has len = {len(parcel)}, unique regions = {len(np.unique(parcel))}', level=1)

    # Apply cifti medial wall reduction to parcellation
    fill_cifti = _remove_medial_wall(fill_cifti, parcel, index_map, _print=_print)

    _print(f'Removed medial wall, new cort unique regions = {len(np.unique(fill_cifti))}', level=1)

    # Add subcortical structures as unique parcels next
    if add_sub:
        fill_cifti = _add_subcortical(fill_cifti, index_map, _print=_print)

    return fill_cifti


def get_cort_slabs(parcel, index_map, _print=None):
    '''Prob. case'''
    
    cort_slabs = []
    for i in range(parcel.shape[1]):
        slab = np.zeros(91282)
        slab = _remove_medial_wall(slab, parcel[:, i], index_map, _print=_print)
        cort_slabs.append(slab)
        
    return cort_slabs
        

def get_subcort_slabs(index_map):
    'Prob. case'
    
    subcort_slabs = []

    for index in index_map:
        if isinstance(index, Cifti2BrainModel):
            if index.surface_number_of_vertices is None:

                slab = np.zeros(91282)
                
                start = index.index_offset 
                end = start + index.index_count
                slab[start:end] = 1
                
                subcort_slabs.append(slab)
                
    return subcort_slabs


def prob_parc_to_cifti(parcel, index_map, _print=None):
    'Prob. case'

    cort_slabs = get_cort_slabs(parcel, index_map, _print=_print)
    subcort_slabs = get_subcort_slabs(index_map)
    
    return np.stack(cort_slabs + subcort_slabs, axis=1)


def surf_parc_to_cifti(cifti_file, parcel, add_sub=True, clean_parc=True, verbose=0):
    '''For now just works when parcel file is a parcellation
    in combined fs_LR_32k lh+rh space with medial wall included.
    Works for static or prob.

    Note: Assumes 0 is background for verbose statements, e.g.,
    when printing info about number of unique regions.

    If clean_parc, then make sure the final parcellations contain
    unique values in order with no missing labels
    '''

    _print = _get_print(verbose=verbose)

    # Get index map from example cifti file
    index_map = nib.load(cifti_file).header.get_index_map(1)

    # Load parcel
    parcel = load(parcel)
    
    # Probabilistic case
    if len(parcel.shape) > 1:
        return prob_parc_to_cifti(parcel, index_map, _print=_print)

    # Static case
    static = _static_parc_to_cifti(parcel=parcel, index_map=index_map,
                                   add_sub=add_sub, _print=_print)
    # Optionally clean labels
    if clean_parc:
        static = clean_parcel_labels(static)

    return static

def add_surface_medial_walls(data):
    '''Right now only works with just concat'ed surface level data.'''
    
    # Load index maps
    if isinstance(data, nib.Cifti2Image):
        index_map = data.header.get_index_map(1)
    else:
        index_map = nib.cifti2.load(data).header.get_index_map(1)

    lh_index, rh_index = index_map[0], index_map[1]
    
    # Make sure right type of cifti, i.e., with medial walls missing
    assert lh_index.surface_number_of_vertices != lh_index.index_count
    
    # Init to fill
    n_vert_per_hemi = lh_index.surface_number_of_vertices
    to_fill = np.zeros(int(n_vert_per_hemi * 2))
    
    # Get separate inds
    lh_inds = np.array(lh_index.vertex_indices._indices)
    rh_inds = np.array(rh_index.vertex_indices._indices)
    
    # Load also as data - okay if already cifti image
    np_data = load(data)
    
    # Fill in
    to_fill[lh_inds] = np_data[:len(lh_inds)]
    to_fill[rh_inds + n_vert_per_hemi] = np_data[len(lh_inds):]
    
    return to_fill

def _get_index_map(data):

    if len(data.shape) == 1:
        cifti_len = data.shape[0]
    elif len(data.shape) == 2:
        cifti_len = data.shape[1]
    else:
        raise RuntimeError('Passed data cannot exceed 2-dimensions!')

    # Hacky solution for now - for when subcort already
    # extracted.
    sub_cifti_mapping = {31870: 91282, 62053: 170494}

    if cifti_len in sub_cifti_mapping:
        cifti_len = sub_cifti_mapping[cifti_len]

    # Get loc if saved
    index_map_loc = os.path.join(data_dr, 'index_maps',
                                 str(int(cifti_len)) + '.pkl')

    if not os.path.exists(index_map_loc):
        raise RuntimeError(f'Cannot find saved cifti index map for cifti length = {cifti_len}!')

    with open(index_map_loc, 'rb') as f:
        index_map = pkl.load(f)

    return index_map

def extract_subcortical_from_cifti(data, index_map=None, space=None):
    '''Given some cifti-data like array and an index_map, 
    extract just the sub-cortical and return as a Nifti1Image
    
    If data is multi-dimensional, assume that the 2nd part of the shape
    corresponds to the cifti-data.
    '''

    from ..plotting.ref import VolRef, _get_vol_ref_from_guess

    # Make sure data has no hanging dimensions
    data = np.squeeze(data)
    
    # Get index map from shape of data if not passed
    if index_map is None:
        index_map = _get_index_map(data)

    # Go through each index map
    sub_mapping, offsets = [], []
    
    for index in index_map:
        if isinstance(index, Cifti2BrainModel):
            if index.surface_number_of_vertices is None:

                # Add ijk mapping
                sub_mapping += index.voxel_indices_ijk._indices

                # Add offsets
                offsets.append(index.index_offset)
    
    # Concat mapping
    sub_mapping = tuple(np.stack(sub_mapping, axis=1))
                
    # Assumes that subcortical are saved at end of cifti array
    index_offset = min(offsets)
    
    # Get vol ref from guess, or if passed
    if space is None:
        vol_ref = _get_vol_ref_from_guess(sub_mapping)
    else:
        vol_ref = VolRef(space=space)
    
    # Fill empty with data based on extracted mapping
    to_fill = np.zeros(vol_ref.shape)
    
    # If doesn't match then try case where no index offset
    # since subcortical already is extracted.
    if len(to_fill[sub_mapping]) != len(data[index_offset:]):
        index_offset = 0

    if len(data.shape) == 1:
        to_fill[sub_mapping] = data[index_offset:]
    elif len(data.shape) == 2:
        to_fill[sub_mapping] = data[:, index_offset:]

    else:
        raise RuntimeError('Passed data cannot exceed 2-dimensions!')

    # Return as Nifti1Image
    return nib.Nifti1Image(to_fill, vol_ref.ref_vol_affine)



'''
Desired future functionality.

- Should be able to add / remove medial wall.
    - Input should be cifti, gifti, or array (i.e., need to have saved data references)
    - Input should be either just lh, just rh, concat surfaces, or full cifti w/ subcortical

- Parcellations should be able to be converted from lh + rh separate or concat,
  to full cifti, note that this isn't possible for normal surfaces. Functionality
  already exists, except should handle some extra cases.

- Would be cool to be able to provide re-sampling between different surfaces

'''


import numpy as np
from math import atan
from scipy.special import erfc, erfcinv, betainc
import numpy as np

def g_to_z(g, df2):

    g = np.squeeze(g)
    
    # Get pos neg index
    idx = g > 0
    
    # Init
    z = np.empty(g.shape)
    
    # Fill pos then negative
    z[idx] = erfcinv(2 * palm_gcdf(-g[idx], df2[idx])) * np.sqrt(2)
    z[~idx] = -erfcinv(2 * palm_gcdf(g[~idx], df2[~idx])) * np.sqrt(2)
    
    return z

def palm_gcdf(g, df2):

    i_c = df2 == 1
    i_n = df2 > 1e7
    i_g = ~(i_c | i_n)
    
    # Init
    gcdf = np.zeros(g.shape)
    
    if i_g.any():

        # Get params
        a, b = df2[i_g] / 2, .5
        
        x = np.real(1 / (1 + np.square(g[i_g]) / df2[i_g]))

        # Apply func
        gcdf[i_g] = betainc(a, b, x) / 2
    
    i_g = g > 0 & i_g
    gcdf[i_g] = 1 - gcdf[i_g]
    
    # Is 1 case
    if i_c.any():
        gcdf[i_c] = .5 + atan(g[i_c]) / np.pi
    
    # Is above 1e7 case
    if i_n.any():
        gcdf[i_n] = erfc(-g[i_n] / np.sqrt(2)) / 2
        
    return gcdf

def freedmanlane(P, Y, Rz, Hz):
    return (P.T @ Rz + Hz) @ Y

def calc_den_fast(eC, cte, r):
    cte_reshape = cte.reshape((r, r, cte.shape[-1])).transpose(2, 0, 1)
    return np.squeeze((eC.T @ np.linalg.pinv(cte_reshape) @ eC).transpose(1, 2, 0))

def fastv(input_matrix, psi, res,
          variance_groups, drm, contrast):

    # Init vars
    r = input_matrix.shape[-1]
    cte  = np.zeros((r ** 2, res.shape[-1]), dtype=input_matrix.dtype.name)

    # For each variance group
    for vg in np.unique(variance_groups):

        # Index of the current vg
        vg_idx = variance_groups == vg

        # Sum the diag resid matrix for this variance group
        diag_resid_sum_vg = np.sum(drm[vg_idx], axis=0)

        # Divide by sum-squared residuals
        weights_vg = diag_resid_sum_vg / np.sum(np.square(res[vg_idx]), axis=0)

         # Get correct piece of input matrix
        input_matrix_vg = input_matrix[vg_idx].T @ input_matrix[vg_idx]

        # Perform matrix dot product between weights_vg and input_matrix_vg
        cte = cte + (np.dot(np.expand_dims(input_matrix_vg.flatten(), axis=1),
                            np.expand_dims(weights_vg, axis=0)))

    # Calc rest of v and return
    return contrast.T @ psi / np.sqrt(
        calc_den_fast(contrast, cte, r))

def fastv_with_df(input_matrix, psi, res, variance_groups, drm, contrast):

    # Init vars
    n_vg = len(np.unique(variance_groups))
    dtype = input_matrix.dtype.name
    nT = res.shape[-1]
    W = np.zeros((n_vg, nT), dtype=dtype)
    r = input_matrix.shape[-1]
    dRmb = np.zeros((n_vg, 1), dtype=dtype)
    cte  = np.zeros((r ** 2, nT), dtype=dtype)

    # For each variance group
    for b, vg in enumerate(np.unique(variance_groups)):

        # Index of the current vg
        vg_idx = variance_groups == vg

        dRmb[b] = np.sum(drm[vg_idx], axis=0)
        W[b] = dRmb[b] / np.sum(np.square(res[vg_idx]), axis=0)
        input_matrix_vg = input_matrix[vg_idx].T @ input_matrix[vg_idx]
        cte = cte + (np.dot(np.expand_dims(input_matrix_vg.flatten(), axis=1), np.expand_dims(W[b], axis=0)))

        W[b] = W[b] * np.sum(vg_idx)

    # Calc den
    den = calc_den_fast(contrast, cte, r)
    G = contrast.T @ psi / np.sqrt(den)

    sW1 = np.sum(W, axis=0)
    bsum = np.sum(np.square((1-(W / sW1))) / dRmb, axis=0)
    df2 = (1/3) / bsum

    return G, df2

def run_permutation(p_set, target_vars, rz, hz,
                    input_matrix, variance_groups,
                    drm, contrast, use_z=False, **kwargs):

    # Make sure passed p_set has same dtype as target vars
    p_set = p_set.astype(target_vars.dtype.name)
    
    # Apply freeman-lane
    Y = freedmanlane(p_set, target_vars, rz, hz)

    # Fit GLM
    psi, _, _, _ = np.linalg.lstsq(input_matrix, Y, rcond=None)
    res = Y - input_matrix @ psi
    del Y

    # Base case, just compute v
    if not use_z:
        v = fastv(input_matrix=input_matrix, psi=psi, res=res,
                variance_groups=variance_groups, drm=drm,
                contrast=contrast)
        return np.squeeze(v)

    # Otherwise, special case, compute z as well
    v, df2 = fastv_with_df(input_matrix=input_matrix, psi=psi, res=res,
                           variance_groups=variance_groups, drm=drm,
                           contrast=contrast)

    # Calculate z from v
    z = g_to_z(v, df2)

    # Return just z instead
    return z




import tensorflow as tf
import tensorflow.experimental.numpy as tnp
import numpy as np

def cast_input_to_tensor(target_vars, rz, hz, input_matrix,
                         drm, contrast, dtype=None):

    return (tf.convert_to_tensor(target_vars, dtype=dtype),
            tf.convert_to_tensor(rz, dtype=dtype),
            tf.convert_to_tensor(hz, dtype=dtype),
            tf.convert_to_tensor(input_matrix, dtype=dtype),
            tf.convert_to_tensor(drm, dtype=dtype),
            tf.convert_to_tensor(contrast, dtype=dtype))

def calc_den_fast(c, cte, r):

    # Reshape and calc pinverse of cte
    cte_inv = tf.linalg.pinv(tf.transpose(tf.reshape(cte,
                                                    (r, r, cte.shape[-1])),
                                          perm=[2, 0, 1]))

    # Apply the contrast to extract correct pieces, and return squeezed
    return tnp.squeeze(tf.transpose((tf.transpose(c) @ cte_inv @ c), perm=[1, 2, 0]))

@tf.function
def proc_vg(vg_idx, drm, res, input_matrix):

    # Sum the diag resid matrix for this variance group
    diag_resid_sum_vg = tnp.sum(drm[vg_idx], axis=0)

    # Divide by sum-squared residuals
    weights_vg = diag_resid_sum_vg / tnp.sum(tnp.square(res[vg_idx]), axis=0)

    # Get correct piece of input matrix
    input_matrix_vg = tf.transpose(input_matrix[vg_idx]) @ input_matrix[vg_idx]

    # Perform matrix dot product between weights_vg and input_matrix_vg
    return tnp.dot(tnp.expand_dims(tf.reshape(input_matrix_vg, [-1]), axis=1),
                   tnp.expand_dims(weights_vg, axis=0))

def fastv(input_matrix, psi, res,
          variance_groups, drm, contrast):
    '''Note, this version of fastv does not calculate df2.'''

    # Init vars
    dtype = input_matrix.dtype.name
    r = input_matrix.shape[-1]
    cte  = tnp.zeros((r ** 2, res.shape[-1]), dtype=dtype)

    # Process each variance group
    for vg in np.unique(variance_groups):

        # Get index of the current vg
        vg_idx = variance_groups == vg
        cte += proc_vg(vg_idx, drm, res, input_matrix)
      
    # Calc rest of v and return
    return tf.transpose(contrast) @ psi / tnp.sqrt(
        calc_den_fast(contrast, cte, r))

def freedmanlane(p_set, target_vars, rz, hz):
    return (tf.transpose(p_set) @ rz + hz) @ target_vars

def run_permutation(p_set, target_vars, rz, hz,
                    input_matrix, variance_groups,
                    drm, contrast, use_z=False, **kwargs):

    if use_z:
        raise RuntimeError('use_z not supported with tensorflow.')

    # Need to convert each passed permutation to Tensor
    # w/ the same type as target
    p_set = tf.convert_to_tensor(p_set, dtype=target_vars.dtype.name)

    # Apply freeman-lane
    targets_permuted = freedmanlane(p_set, target_vars, rz, hz)
   
    # Fit GLM
    psi = tf.linalg.lstsq(input_matrix, targets_permuted, fast=True)
    res = targets_permuted - input_matrix @ psi
    del targets_permuted

    # Compute v stat
    v = fastv(input_matrix=input_matrix, psi=psi, res=res,
              variance_groups=variance_groups, drm=drm,
              contrast=contrast)
    
    # Return as numpy array for max compat
    return np.array(tnp.squeeze(v))


import numpy as np
from ..loading import load

def _proc_background_label(data, background_label):
    '''Sets passed background label to 0'''

    # If None or 0, do nothing
    if background_label is None or background_label == 0:
        return data

    # If list / array-like
    if hasattr(background_label, '__iter__'):
        data[np.isin(data, background_label)] = 0

    # If non-zero single label
    else:
        data[data == background_label] = 0

    return data

def merge_parc_hemis(lh, rh, background_label=0, merge_thresh=.75, clean_parc=True):
    '''This function is designed to help merge two surface parcellations when
    saved in separate files or data arrays.

    Parameters
    -----------
    lh : str, :func:`arrays<numpy.array>` or other
        The str location or loaded numpy array-like representing
        the left hemisphere values in which to merge.

    rh : str, :func:`arrays<numpy.array>` or other
        The str location or loaded numpy array-like representing
        the right hemisphere values in which to merge.

    background_label : int, list-like of int or None, optional
        Optionally may pass a value, or list-like of values in
        which if found in either the left hemisphere or
        right hemisphere parcellation data will be set to 0.

        If passed 0 or None, will ignore.

        ::

            default = 0

    merge_thresh : float between 0 and 1, optional
         This function will handle both cases where the values
        between each file are separate (e.g., 0 to n-1 in each) or
        overlapping. The way this method distinguishes between
        these cases is by looking at the percent of labels which
        intersect between parcellation hemispheres. Specifically,
        if the number of intersecting values / the number of left hemisphere
        unique values is greater than this `merge_thresh`, the values
        in the right hemisphere data will be incremented such that they
        are treated as unique values.

        ::

            default = .75

    clean_parc : bool, optional
        If True, ensures that labels are set to be 0 to n-1
        removing any labels that are not sequential.

        ::

            default = True
    '''

    assert merge_thresh >= 0, 'Cannot be negative.'
    assert merge_thresh <= 1, 'Cannot be above 1.'

    # Load if needed
    lh, rh = load(lh), load(rh)

    # Process background label (set to 0 if any)
    lh = _proc_background_label(lh, background_label)
    rh = _proc_background_label(rh, background_label)
    
    # Get unique by hemisphere
    ul, ur = np.unique(lh), np.unique(rh)
    
    # Calculate intersection
    intersect = len(np.intersect1d(ul, ur))
    
    # If over 75% of labels intersect, merge
    # by adding other hemi as unique
    if intersect > 0:
        if (intersect / len(ul)) > merge_thresh:

            # Add max + 1, since parcs can start at 0
            rh[rh != 0] += (max(ul) + 1)

    # Otherwise, just concatenate as is
    data = np.concatenate([lh, rh])

    # Optionally clean parc labels
    if clean_parc:
        data = clean_parcel_labels(data)

    return data

def clean_parcel_labels(parcel):
    
    # Get unique regions
    u_regions = np.unique(parcel)

    # If missing 0, pretend there
    u_regions = np.union1d([0], u_regions)

    # Should be sorted, but make sure sorted
    u_regions = np.sort(u_regions)
    
    # If already 'clean' return as is
    clean_regions = np.arange(len(u_regions))
    if len(np.setdiff1d(clean_regions, u_regions)) == 0:
        return parcel
    
    # Fill in with new clean values
    clean_parcel = np.zeros(parcel.shape, dtype=parcel.dtype)
    for clean, original in zip(np.arange(len(u_regions)), u_regions):
        clean_parcel[parcel == original] = clean
        
    return clean_parcel


from ..loading.from_data import get_surf_loc, _load_medial_wall
from ..loading.funcs import load
from ..transform.space import _resolve_space_name
import warnings

def vol_to_surf(vol, target_space='32k_fs_LR',
                method='auto', ref_key='pial',
                **kwargs):
    '''

    The ref argument is only used with the nilearn style projection.

    If using the nilearn style vol to surf projection, the default
    values which can be overridden via kwargs are:

    radius=3.0, interpolation='linear', kind='auto',
    n_samples=None, mask_img=None,
    inner_mesh=None, depth=None.
    
    
    '''
    # Make sure valid target space
    target_space = _resolve_space_name(target_space)

    if method not in ['auto', 'nilearn', 'hcp']:
        raise RuntimeError('method must be one of auto, nilearn, hcp.')

    # Auto just tries hcp with tolerance for failure
    if method == 'auto':

        try:
            return _hcp_vol_to_surf(vol, target_space)
        except:
            print('Using nilearn resample.')

    # Only different is if error and hcp requested
    # pass along error.
    elif method == 'hcp':
        return _hcp_vol_to_surf(vol, target_space)
    
    # Base case is nilearn
    return _nilearn_vol_to_surf(vol, target_space, ref_key, **kwargs)
    

def _nilearn_vol_to_surf(vol, target_space, ref_key, **kwargs):

    with warnings.catch_warnings():
        warnings.simplefilter(action='ignore', category=FutureWarning)
        from nilearn.surface import vol_to_surf

    projected = {}
    for hemi in ['lh', 'rh']:
        surf_mesh = get_surf_loc(space=target_space,
                                 hemi=hemi, key=ref_key)
        projected[hemi] = vol_to_surf(vol, surf_mesh=surf_mesh, **kwargs)

    return projected

def _get_fs_density(target_space):

    sz = len(_load_medial_wall(target_space))

    try:
         from neuromaps import transforms
    except ImportError:
        raise ImportError('Download neuromaps to use this method with pip install https://github.com/netneurolab/neuromaps/archive/refs/tags/0.0.1.zip')

    # Try to get density from base library
    try:
        return transforms.DENSITY_MAP[sz]
    except KeyError:
        raise RuntimeError(f'target_space: {target_space} not supported with hcp style resampling.')

def _hcp_vol_to_surf(vol, target_space):

    try:
         from neuromaps import transforms
    except ImportError:
        raise ImportError('Download neuromaps to use this method with pip install https://github.com/netneurolab/neuromaps/archive/refs/tags/0.0.1.zip')

    if 'fs_LR' in target_space:
        ret = transforms.mni152_to_fslr(vol, fslr_density=target_space.split('_')[0])
    elif 'fsaverage' in target_space:
        ret = transforms.mni152_to_fsaverage(vol, fsavg_density=_get_fs_density(target_space))
    elif 'civet' in target_space:
        ret = transforms.mni152_to_civet(vol, civet_density='41k')
    else:
        raise RuntimeError(f'target_space: {target_space} not supported with hcp style resampling.')

    # Convert return to loaded style
    return {'lh': load(ret[0]), 'rh': load(ret[1])}



from .conv import extract_subcortical_from_cifti
import nibabel as nib
import os
import pandas as pd
import numpy as np
from ..loading.funcs import load
from ..loading.from_data import _load_medial_wall, auto_load_rois
from ..loading.space import get_space_options
from .. import data_dr
from ..misc.print import _get_print

def _resolve_space_name(space):

    # Base
    if space is None:
        return None
    if space == 'native':
        return 'native'

    # Check if already okay
    space_options = os.listdir(data_dr)
    if space in space_options:
        return space

    # If not, let's let some things slide
    if space == '41k_civet':
        return 'civet'

    # Make some replacements
    space = space.replace('Civet', 'civet')
    space = space.replace('_lr', '_LR')
    space = space.replace('_FS_', '_fs_')

    # Check again
    if space in space_options:
        return space

    raise RuntimeError('Invalid space passed!')

def _load(data):
    '''Simple wrapper around load with
    the added attribute that if a str location is passed
    we want to only if the data is multi-dimensional - try to
    load it as a Nifti1Image, i.e., with affine, otherwise
    perform normal load as array.'''
    
    # If already nifti image, keep as is
    if isinstance(data, nib.Nifti1Image):
        return data
    
    # First, perform normal load
    loaded_data = load(data)
   
    # If not str, i.e., already passed as obj
    # return as is
    if not isinstance(data, str):
        return loaded_data
    
    # If loaded data is flat array, return as is
    if len(loaded_data.shape) == 1:
        return loaded_data

    # Last case is try to load instead w/ nibabel load
    # if fails, return flat array
    try:
        return nib.load(data)
    except:
        return loaded_data

def _get_space_mapping(hemi):

    # Get options for spaces
    space_options = get_space_options()
    mapping = {}

    # For each potential space
    for space in space_options:

        # See if medial wall, skip if isn't
        try:
            medial_wall = _load_medial_wall(space, hemi=hemi)
        except RuntimeError:
            continue
        
        # Add base size first
        base_sz = len(medial_wall)

        # Check if already exists
        if base_sz in mapping:

            # In case of duplicates, favor civet, then fsaverage, by skipping
            # adding new, and keeping current
            # Note, potential TODO may need more cases in future
            if 'civet' in mapping[base_sz][0]:
                continue
            if 'fsaverage' in mapping[base_sz][0]:
                continue

        # If here, then add, replacing if already there
        mapping[base_sz] = (space, False)

        # Add no medial wall version
        no_medial_sz = int(np.sum(medial_wall))
        if no_medial_sz in mapping:
            # If this happens will break later logic
            raise RuntimeError('Should not be overlaps here.')
        mapping[no_medial_sz] = (space, True)

    return mapping

def _check_rois(data):

    # The idea here is that if data is passed
    # as a pandas Series or DataFrame, then definetly
    # ROIs
    if isinstance(data, (pd.Series, pd.DataFrame)):
        return True

    # If dict ... check to make sure not space dict
    if isinstance(data, dict):
        bad_keys = ['lh', 'rh', 'concat', 'vol', 'sub']
        for key in bad_keys:
            if key in data:
                return False

        return True

    # Return false otherwise
    return False

def process_space(data, space=None, hemi=None,
                  verbose=1, _print=None):
    '''
    If hemi=None, auto-detect. Or can pass as
    lh or rh, as this is the only case
    where it could go either way arbitrarily.
    
    Will return packed data in a tuple if lh+rh or
    lh+rh+sub.

    - Process passed sub-cortical to make sure in nifti
    - If surface data is missing medial wall, add it here

    Space argument refers to the space of the surface
    data. 
    

    Data can be passed as
    
    1. A single data representing surf, surf+surf, surf+surf+vol or just vol
    2. A single file location w/ a data array
    3. A list / array-like of length either 2 or 3, if 2 then represents surf+surf
       if 3 then surf+surf+vol
    4. A list / array-like same as above, but with file-paths
    '''

    # Resolve user passed space
    space = _resolve_space_name(space)

    _print = _get_print(verbose, _print)
    _print('Calling function process_space on data.', level=2)
    
    # Proc hemi input
    if hemi == 'left':
        hemi = 'lh'
    if hemi == 'right':
        hemi = 'rh'
    assert hemi in ['lh', 'rh', None], 'hemi must be lh, rh or None!'

    # Check to see if the data was passed as a list of ROIs + values
    # or atleast try ... 
    is_rois = _check_rois(data)
    if is_rois:
        return auto_load_rois(data, space=space, hemi=hemi, _print=_print)
    
    # Space info- sep by hemi
    lh_space_mapping = _get_space_mapping(hemi='lh')
    rh_space_mapping = _get_space_mapping(hemi='rh')
    
    # Gen combinations of hemi sizes
    hemi_sizes = []
    for lk, rk in zip(lh_space_mapping, rh_space_mapping):
        hemi_sizes.append((lk, rk))

    # TODO add cases for when just subcortical / volumetric is passed
    # Or flattened voxel or something
    # not in cifti form, but in flattened sub-cortical voxel form
    # For now these are just from cifti...
    sub_sizes = [31870, 62053]

    # Init.
    proc_data = {}
    lh_data, rh_data, vol_data = None, None, None

    # If data is passed as str - replace with loaded
    # version here - before main loop with cases
    if isinstance(data, str):
        _print(f'Loading data from location: {data}.', level=2)
        data = _load(data)

    elif isinstance(data, nib.Cifti2Image):
        # @TODO if cifti use this info directly as it contains
        # space information and whatnot. For now, just treat as array
        _print(f'Loading data from cifti as data array.', level=2)
        data = _load(data)

    # Gifti case directly
    elif isinstance(data, nib.GiftiImage):
        _print(f'Loading data from gifti as data array.', level=2)
        data = _load(data)
        
    # Otherwise, if data is already in dict form
    # check from there
    if isinstance(data, dict):

        if 'lh' in data:
            lh_data = _load(data['lh'])
        if 'rh' in data:
            rh_data = _load(data['rh'])
        if 'concat' in data:
            sz = len(data['concat']) // 2
            lh_data, rh_data = _load(data['concat'][:sz]), _load(data['concat'][sz:])

        if 'vol' in data:
            vol_data = _load(data['vol'])
        
        # Also support passing as sub
        if 'sub' in data:
            vol_data = _load(data['sub'])

    # Passed just single subcort case / volumetric
    elif isinstance(data, nib.Nifti1Image):
        vol_data = data

    # Check if
    # was passed as a list / array-like corresponding
    # to lh+rh data
    elif len(data) == 2:
        _print('Assuming data passed as [lh, rh].', level=1)
        lh_data, rh_data = _load(data[0]), _load(data[1])

    # Or lh+rh+sub case
    elif len(data) == 3:
        _print('Assuming data passed as [lh, rh, sub].', level=1)
        lh_data, rh_data, vol_data = _load(data[0]), _load(data[1]), _load(data[2])

    # Single passed array case
    # Since data is already array if here
    else:
        _print('data was passed as single array, automatically inferring splits / space.', level=2)

        for lh_sz, rh_sz in hemi_sizes:
            
            # If matches single hemi length
            if len(data) == lh_sz:
                
                # If also matches rh, and hemi is rh, override
                if len(data) == rh_sz and hemi == 'rh':
                    rh_data = data

                # Otherwise all other cases + default
                # are lh
                else:
                    lh_data = data
                break

            # Check if it matches a concat case
            elif len(data) == lh_sz + rh_sz:
                lh_data, rh_data = data[:lh_sz], data[lh_sz:]
                break

            # Go through sub cases
            for sub_sz in sub_sizes:

                # lh+rh+sub case
                if len(data) == lh_sz + rh_sz + sub_sz:
                    lh_data, rh_data = data[:lh_sz], data[lh_sz:lh_sz+rh_sz]
                    vol_data = data[lh_sz+rh_sz:]
                    break

                # Just sub case
                elif len(data) == sub_sz:
                    vol_data = data
                    break

    # Make sure something found, if not error
    if lh_data is None and rh_data is None and vol_data is None:
        raise RuntimeError(f'Could not determine the space / hemi of passed data with len={len(data)}')
    
    # Process found data - don't need
    # user specified space here as still need auto-detect info
    # for if need proc medial walls.
    lh_space, rh_space = None, None
    if lh_data is not None:
        proc_data['lh'], lh_space = proc_hemi_data(lh_data, 'lh', lh_space_mapping)
    if rh_data is not None:
        proc_data['rh'], rh_space = proc_hemi_data(rh_data, 'rh', rh_space_mapping)
    if vol_data is not None:
        proc_data['vol'] = proc_vol_data(vol_data)

    _print(f'Inferred passed data: {list(proc_data)}', level=1)

    # Process space combinations
    detected_space = None
    if lh_space is None and rh_space is None:
        if 'vol' not in proc_data:
            raise RuntimeError('No valid data found')
    elif lh_space is not None and rh_space is None:
        detected_space = lh_space
    elif lh_space is None and rh_space is not None:
        detected_space = rh_space
    elif lh_space is not None and rh_space is not None:
        assert lh_space == rh_space, 'Auto detected spaces do not match!'
        detected_space = lh_space
    
    # If user passed space, use that instead
    if space is not None:
        detected_space = space
    else:
        _print(f'Inferred surface space: {detected_space}', level=1)

    # Return proc data + space
    return proc_data, detected_space

def proc_hemi_data(hemi_data, hemi, space_mapping):
    
    # If doesn't match any, assume in native space
    if len(hemi_data) not in space_mapping:
        return hemi_data, 'native'
    
    # Check space mapping dict
    space, needs_medial_wall = space_mapping[len(hemi_data)]
    
    # Add medial wall if needs it
    if needs_medial_wall:
        medial_wall_mask = _load_medial_wall(space, hemi)
        to_fill = np.zeros(medial_wall_mask.shape)

        # Medial wall saved as 0 in mask so fill in hemi data like this
        to_fill[medial_wall_mask] = hemi_data

        return to_fill, space
    
    # If already okay, return as is, with space
    return hemi_data, space

def proc_vol_data(vol_data):
    
    # If already nifti
    if isinstance(vol_data, nib.Nifti1Image):
        return vol_data

    # Otherwise, if size for cifti rois
    if len(vol_data) in [31870, 62053]:
        vol_data = extract_subcortical_from_cifti(vol_data)
    
    # The other case will be flattened subcort voxel case
    else:
        raise RuntimeError('Not implemented.')

    return vol_data



import os
import warnings
import numpy as np
from scipy.stats import ks_2samp, gaussian_kde
from scipy.spatial.distance import jensenshannon
from .parc import merge_parc_hemis

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import MinMaxScaler
import itertools
import pandas as pd

from ..loading import load
from ..stats.orth import OrthRegression
from scipy.stats import pearsonr

def _get_kde_p(data, n=256):
    kde = gaussian_kde(data)
    return kde(np.linspace(np.min(data), np.max(data), n))

def _jsd_metric(x, y):

    # Get kde est
    x_p = _get_kde_p(x)
    y_p = _get_kde_p(y)

    # Then return 1 - dist
    return 1 - jensenshannon(x_p, y_p)

def _ks_metric(x,  y):
    '''1 - kd_2samp metric'''

    return 1 - ks_2samp(x, y).statistic

def _ks_demean_metric(x, y):
    
    return _ks_metric(x - np.mean(x),
                      y - np.mean(y))

def _ks_normalize_metric(x, y):
    
    return _ks_metric((x - np.mean(x)) / np.std(x),
                      (y - np.mean(y)) / np.std(y))

def gen_indv_roi_network(data, labels, metric='jsd', vectorize=False, discard_diagonal=False):
    '''This function is designed to generate a network
    of 1 - distance function between groups of different ROIs.
    

    Specifically, this method calls the function :func:`scipy.stats.ks_2samp`  or
    :func:`scipy.spatial.distance.jensenshannon` to calculate the distances
    between each collection of data points in each pair of ROIs, calculating the
    distance between the two distributions. Values are then subtracted from 1,
    such that exact distribution matches have value 1, i.e., these are the
    strongest connections. 

    Parameters
    ----------
    data : 1D numpy array
        Data at this stage must be a single dimensional
        numpy array representing the underlying data in which
        the labelled regions correspond to.

        For example, this is typically a subject's neuroimaging
        data along with corresponding label file in some ROI space.

    labels : 1D numpy array
        Labels in the form of a 1D numpy array with the same shape / len
        as data. This file should contain integer labels
        corresponding to which vertex or data elements belong
        to different ROIs.

    metric : {'jsd', 'ks', 'ks_demean', 'ks_normalize'}, optional
        The type of distance to compute between points.

        - 'jsd' : Use the 1 - Jensen Shannon distance between each rois kde estimated distributions.
        - 'ks' : Use the 1 - Kolmogorov–Smirnov distance between distributions.
        - 'ks_demean' : Same as 'ks', but with each ROI's points de-meaned.
        - 'ks_normalize' : Same as 'ks', but with each ROI's points normalized.

        ::

            default = 'jsd'

    vectorize : bool, optional
        If True, matrices are reshaped into 1D arrays and only their
        flattened lower triangular parts are returned.

        ::

            default = False

    discard_diagonal : bool, optional
        If True, the diagonal elements of the calculated matrix are
        discarded.

        ::

            default = False

    Returns
    --------
    matrix : 1/2D numpy array
        Base behavior is a two dimensional array containing the
        distances / statistics between
        every combination of ROI. If vectorize is True, then
        a 1D array is returned instead, and if discard_diagonal
        is set, then the diag will be removed as well.
    '''

    # Get correct metric function
    if metric == 'jsd':
        dist_metric = _jsd_metric
    elif metric == 'ks':
        dist_metric = _ks_metric
    elif metric == 'ks_demean':
        dist_metric = _ks_demean_metric
    elif metric == 'ks_normalize':
        dist_metric = _ks_normalize_metric
    else:
        raise RuntimeError(f'Passed metric {metric} invalid!')

    # Warn if improper input
    if discard_diagonal is True and vectorize is False:
        warnings.warn('Note: not discarding diagonal since vectorize is False.')

    # Generate the corr matrix
    n_unique = len(np.unique(labels))
    matrix = np.zeros((n_unique, n_unique))

    for i, u1 in enumerate(np.unique(labels)):
        for j, u2 in enumerate(np.unique(labels)):

            # Skip self
            if i == j:
                continue

            # If already calculated reverse, skip
            if matrix[j][i] == 0:
                matrix[i][j] = dist_metric(data[labels==u1], data[labels==u2])
                matrix[j][i] = matrix[i][j]

    # If vectorize, return just lower triangle
    if vectorize:

        # Keep diagonal or not
        k = 0
        if discard_diagonal:
            k = -1

        return matrix[np.tril_indices(matrix.shape[0], k=k)]

    # Otherwise, return full matrix
    return matrix

def _load_fs_subj_data(subj_dr, modality, parc):
    
    # Path to their thickness file
    data_lh = load(os.path.join(subj_dr, 'surf', f'lh.{modality}'))
    data_rh = load(os.path.join(subj_dr, 'surf', f'rh.{modality}'))
    data = np.concatenate([data_lh, data_rh])

    # Get merged label file
    lh_labels = os.path.join(subj_dr, 'label', f'lh.{parc}')
    rh_labels = os.path.join(subj_dr, 'label', f'rh.{parc}')
    labels = merge_parc_hemis(lh_labels, rh_labels)
    
    return data, labels

def gen_fs_subj_vertex_network(subj_dr, modality='thickness',
                               parc='aparc.a2009s.annot', metric='jsd',
                               vectorize=False, discard_diagonal=False):
    '''This function is helper function for calling
    :func:`gen_indv_roi_network`, for data organized in
    freesurfer individual subject directory style.

    Parameters
    ----------
    subj_dr : str
        The str location of the subject's freesurfer
        directory in which to generate the ks roi network.

    modality : str, optional
        The name of the modality (e.g., thickness, sulc, area, ...)
        in which to generate the network. Should
        be saved in subdirectory surf with names:
        `lh.{modality}` and `rh.{modality}`

        ::

            default = 'thickness'

    parc : str, optional
        The name of the label / parcellation file in which
        to use, as found in subdirectory label with names:
        `lh.{parc}` and `rh.{parc}`.

        These are concatenated internally with function
        :func:`merge_parc_hemis`.

        The default value is to use the destr. parcellation.

        ::

            default = 'aparc.a2009s.annot'

     metric : {'jsd', 'ks', 'ks_demean', 'ks_normalize'}, optional
        The type of distance to compute between points.

        - 'jsd' : Use the 1 - Jensen Shannon distance between each rois kde estimated distributions.
        - 'ks' : Use the 1 - Kolmogorov–Smirnov distance between distributions.
        - 'ks_demean' : Same as 'ks', but with each ROI's points de-meaned.
        - 'ks_normalize' : Same as 'ks', but with each ROI's points normalized.

        ::

            default = 'jsd'

    vectorize : bool, optional
        If True, matrices are reshaped into 1D arrays and only their
        flattened lower triangular parts are returned.

        ::

            default = False

    discard_diagonal : bool, optional
        If True, the diagonal elements of the calculated matrix are
        discarded.

        ::

            default = False

    Returns
    --------
    matrix : 1/2D numpy array
        Two dimensional array containing the
        1 - Kolmogorov–Smirnov statistic / distances between
        every combination of ROI.

        If vectorize is True, then a 1D array is returned.

    '''
    
    # Load in data
    data, labels = _load_fs_subj_data(subj_dr, modality, parc)

    # Call main function
    return gen_indv_roi_network(data, labels, metric=metric,
                                vectorize=vectorize,
                                discard_diagonal=discard_diagonal)


class GroupDifNetwork(BaseEstimator, TransformerMixin):

    _needs_fit_index = True
    
    def __init__(self, vectorize=True, scale_weights=False,
                 fit_only_group_index=None, pval_thresh=None,
                 passthrough=False, verbose=0):
        
        self.vectorize = vectorize
        self.scale_weights = scale_weights
        self.fit_only_group_index = fit_only_group_index
        self.pval_thresh = pval_thresh
        self.passthrough = passthrough
        self.verbose = verbose

    def proc_X(self, X, fit_index=None):
        
        # Use method from sklearn base estimator
        self._check_feature_names(X, reset=True)

        # If df, get fit_index, but convert to numpy array
        # internally, so easier to work with, i.e., consistent indexing.
        if isinstance(X, pd.DataFrame):
            fit_index = X.index
            X = np.array(X)

        return X, fit_index

    def fit(self, X, y=None, fit_index=None):

        X, fit_index  = self.proc_X(X, fit_index=fit_index)

        # More efficient to call fit transform
        # directly, but should still support calling
        # separate

        # Need to call fit transform
        # for fit, if scale weights is True
        if self.scale_weights:
            self.fit_transform(X=X, y=y, fit_index=fit_index)
        
        # Otherwise can just call base fit
        else:
            self._fit(X=X, y=y, fit_index=fit_index)
        
        return self
        
    def _fit(self, X, y=None, fit_index=None):

        if self.verbose > 1:
            print(f'Passed X shape={X.shape} to fit')

        if self.fit_only_group_index is not None:

            if fit_index is None:
                raise RuntimeError('Since using fit_only_group_index, need fit_index here!')

            # Set inds to only group
            inds = fit_index.isin(self.fit_only_group_index)
            X_s = X[inds]
        
        # Otherwise use all subjects to fit reference group
        else:
            X_s = X

        if self.verbose > 1:
            print('Fitting with shape:', X_s.shape, flush=True)
        
        # Gen each possible pair        
        self.pairs_ = list(itertools.combinations(range(X.shape[1]), 2))

        # Fit orth regression for each combination of regions
        self.estimators_ = []
        for i, j in self.pairs_:

            # If non-null pval thresh, add also if
            # base pearson pval is over threshold, add to skip mask
            if self.pval_thresh is not None:
                corr_sig = pearsonr(X_s[:, i], X_s[:, j])[1]

                # If doesn't pass, skip
                if corr_sig > self.pval_thresh:
                    self.estimators_.append(None)
                    continue
            
            # Otherwise, fit OrthRegression
            model = OrthRegression().fit(X_s[:, i], X_s[:, j])
            self.estimators_.append(model)

        
        return self

    def _proc_scale(self, X_trans, fit):

        # If scale weights is True
        if self.scale_weights:
            
            # Fit case, init scaler and fit
            if fit:

                # Fit to X_trans between 0 and 1
                self.scaler_ = MinMaxScaler(feature_range=(0, 1), clip=True)
                self.scaler_.fit(X_trans)
            
            # Always transform
            X_trans = self.scaler_.transform(X_trans)

        return X_trans
    
    def _transform(self, X, fit=False):
        
        # Go through each fitted, and get the residuals
        X_trans = []
        for estimator, ij in zip(self.estimators_, self.pairs_):

            # Just add as NaNs for all if skip
            if estimator is None:
                nans = np.empty(len(X))
                nans[:] = np.nan
                X_trans.append(nans)
                continue
            
            # Unpack
            i, j = ij

            # Get the orth distance between estimate
            # and current point - this estimate is symmetric
            X_trans.append(estimator.get_orth_dist(X[:, i], X[:, j]))
        
        # Set as # of subjects x number of edge combos
        X_trans = np.stack(X_trans, axis=-1)

        # Process if scale weights
        X_trans = self._proc_scale(X_trans, fit)

        if self.verbose > 1:
            print(f'Base transformed shape: {X_trans.shape}', flush=True)
                
        # Return as is, if default
        if self.vectorize:

            # If passthrough and vectorize, then combine
            if self.passthrough:
                return np.hstack([X_trans, X])

            return X_trans
        
        # Otherwise, cast X_trans to full network matrix
        return self._to_matrix(X, X_trans)
    
    def transform(self, X):
 
        # Base transform w/o fit
        return self._transform(X, fit=False)
    
    def _to_matrix(self, X, X_trans):
        '''Transforms to full matrix, w/ redundancy'''
        
        # Init empty network to fill, where 0's represent
        # same pair connections
        as_matrix = np.zeros((X.shape[0], X.shape[1], X.shape[1]))
        for feat, ij in enumerate(self.pairs_):
            i, j = ij
            as_matrix[:, i, j] = X_trans[:, feat]
            as_matrix[:, j, i] = X_trans[:, feat]
            
        return as_matrix
    
    def fit_transform(self, X, y=None, fit_index=None):

        X, fit_index  = self.proc_X(X, fit_index=fit_index)
        
        # Fit transform, calling base fit and base transform _transform
        return self._fit(X=X, y=y, fit_index=fit_index)._transform(X=X, fit=True)





from nilearn.connectome.connectivity_matrices import vec_to_sym_matrix
import numpy as np
import matplotlib.pyplot as plt
import networkx as nx
from sklearn.preprocessing import MinMaxScaler
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec, GridSpecFromSubplotSpec
import nibabel as nib
from ..transform.rois import project_map_fis
from .plot import plot

from matplotlib.cm import get_cmap
from matplotlib.colors import Normalize
import seaborn as sns
from .plot import _proc_vs, _get_if_sym_cbar, _proc_cmap

from textwrap import wrap

def get_colors(weights, ref_weights, symmetric_cbar='auto',
               cmap='default', vmin=None, vmax=None):

    # Idea is use ref weights to determine sym cbar, cmap, vmin, vmax
    # then normalize just the base weights - since get colors is used on a subset of
    # either the degrees or edges ... 

    # Process automatic symmetric_cbar
    symmetric_cbar = _get_if_sym_cbar(ref_weights,
                                      symmetric_cbar=symmetric_cbar,
                                      rois=False) 

    # Get cmap based on passed + if rois or not + if sym
    cmap = _proc_cmap(cmap, rois=False, symmetric_cbar=symmetric_cbar, flat_data=ref_weights)
    our_cmap = get_cmap(cmap)
    
    # Get normalize object - should end up 0 to 1
    vmin, vmax = _proc_vs(ref_weights, vmin=vmin, vmax=vmax,
                          symmetric_cbar=symmetric_cbar)
    norm = Normalize(vmin=vmin, vmax=vmax)
    
    return [our_cmap(norm(w)) for w in weights]

def _scale_weights(weights, ref_weights):
    
    # If includes negative weights, we want to scale
    # in a way that preserves positive and negative ... 
    # But we also don't want to assume that the vmax is the same as the vmin.
    # Solution here is to normalize the absolute weights, then re-cast to original sign
    # If all positive to begin with, no harm done
    signs = np.sign(weights)
    ref_signs = np.sign(ref_weights)

    # Fit min max scaler on ref weights, then apply to base weights
    # so scaled weights between 0 and 1, based on all edges
    scaler = MinMaxScaler().fit(np.abs(ref_weights).reshape((-1, 1)))
    scaled_weights = scaler.transform(np.abs(weights).reshape((-1, 1))).reshape(weights.shape)
    ref_scaled_weights = scaler.transform(np.abs(ref_weights).reshape((-1, 1))).reshape(ref_weights.shape)

    # Re-apply signs
    scaled_weights *= signs
    ref_scaled_weights *= ref_signs

    return scaled_weights, ref_scaled_weights

def add_subgraph_plot(sub_G, G, ax, edge_scale=5,
                      layout='spring', x_scale=.8,
                      y_scale=1, color_edges=True,
                      min_node_sz=200, max_node_sz=400):
    
    # Use weights, + ref for edge thickness
    weights = np.array(list(nx.get_edge_attributes(sub_G, 'weight').values()))
    ref_weights = np.array(list(nx.get_edge_attributes(G, 'weight').values()))

    # Get scaled 0-1 (or -1, 1) weights + ref scaled weight w/ signs preserved
    scaled_weights, ref_scaled_weights = _scale_weights(weights, ref_weights)

    # Optionally color edges
    if color_edges:
        colors = get_colors(scaled_weights, ref_scaled_weights)
        node_color = 'grey'
    else:
        colors = None
        node_color = None

    # Multiply further by edge, after weights potentially used by color
    scaled_weights = scaled_weights * edge_scale
    
    # Set pos by layout choice
    if layout == 'spring':
        pos = nx.spring_layout(sub_G)
    elif layout == 'circular':
        pos = nx.circular_layout(sub_G)

    # Custom
    else:
        pos = layout(sub_G)

    # Size nodes just by their weighted degree in the full graph
    # Make sure to take absolute - Should abs be before degree?
    # Or maybe it is better to just correlate to the listed weighted degree
    ref_degrees = nx.degree(G, weight='weight')
    degs = [np.abs(ref_degrees[node]) for node in sub_G.nodes()]
    ref_degs = np.abs([x[1] for x in list(ref_degrees)])

    # Init normalize object on the reference degree dist from full G
    # Then apply to just the degrees being plotted in this sub-graph,
    # as further sized by passed constraints
    norm = Normalize(np.min(ref_degs), np.max(ref_degs))
    sizes = (norm(degs).data * (max_node_sz - min_node_sz)) + min_node_sz

    # Plot
    nx.draw(sub_G, pos,
            node_color=node_color,
            node_size=sizes,
            edge_color=colors,
            width=scaled_weights,
            with_labels=True, ax=ax)

    # Scale based on passed params to fix out of bounds labels
    ax.set_xlim([x_scale*x for x in ax.get_xlim()])
    ax.set_ylim([y_scale*y for y in ax.get_ylim()])

def _get_diag(x):
    
    around = int(np.sqrt((x * 2)))
    
    for d_len in range(max([around, around - 5]), around+5):
        
        if x == int(((d_len * d_len) - d_len) / 2):
            return np.zeros(d_len)
        elif x == (d_len * d_len) / 2:
            return None
        
def to_sym(betas):
    '''Assumes betas is numpy array'''
    
    return vec_to_sym_matrix(betas, diagonal=_get_diag(len(betas)))

def _beta_to_G(betas, label_names, n_std=None):
    
    if n_std is None:
        thresh = np.min(betas) - 1
    else:
        thresh = np.mean(betas) + (np.std(betas) * n_std)

    # Make graph
    G = nx.Graph()
    for i in range(len(betas)):
        for j in range(len(betas[i])):
            if betas[i][j] > thresh:
                G.add_edge(label_names[i], label_names[j], weight=betas[i][j])
                
    return G

def _line(fig, gs, i):
    
    line_ax = fig.add_subplot(gs[i, :])
    line_ax.axhline(y=1, ls='--', color='gray')
    line_ax.set_axis_off()
    line_ax.patch.set_alpha(0)

def _plot_brain(data, fig, ax, ref, label_names,
                show_abs=True, is_top=False, colorbar=True,
                threshold='auto', **plot_kwargs):

    # For now, if ref is None, let's assume
    # we are plotting just auto roi names to surface
    # TODO will need to ass more cases
    if ref is None:
        if is_top:
            b_sz = .25
            wspace = 0
        else:
            b_sz = .1
            wspace = .1

        # Call plot
        plot(data,
             view='flat standard', b_sz=b_sz, wspace=wspace,
             figure=fig, ax=ax, colorbar=colorbar, aspect=10,
             shrink=.75, threshold=threshold, **plot_kwargs)

    # 4D prob. case
    elif isinstance(ref, nib.Nifti1Image) and len(ref.shape) == 4:

        if is_top:
            display_mode = 'lyrz'
        else:
            display_mode = 'ortho'

        # Project onto data
        degree_proj = project_map_fis(data, ref, label_names)

        # Then plot in axes
        plot(degree_proj, display_mode=display_mode, axes=ax,
             plot_abs=show_abs, threshold=threshold,
             colorbar=colorbar, **plot_kwargs)

def _add_top_figure(degrees, gs, fig, ref,
                    label_names, show_abs=True, **plot_kwargs):

    # For very top row, / the full degree projection
    # Use sub gridspec for more spacing control
    sub_gs = GridSpecFromSubplotSpec(nrows=1, ncols=3,
                                     subplot_spec=gs[0, :],
                                     width_ratios=[.2,  1, .2],)
    full_ax = fig.add_subplot(sub_gs[1])

    # Make brain plot
    _plot_brain(degrees, fig, full_ax, ref, label_names,
                show_abs=show_abs, is_top=True, colorbar=True, **plot_kwargs)

    # Add title + line
    full_ax.set_title('Projection of All Weighted Degrees', size=24, y=1.05)
    _line(fig, gs, 1)

def _plot_ref(ref, node, top_feats, gs, i,
              label_names, fig, show_abs, **plot_kwargs):
    
    # Gen ref values for plotting roi
    if ref is None:

        # Handle pos / neg only case
        if top_feats[node] < 0:
            add = 1
        else:
            add = -1

        roi_vals = {k: top_feats[node] + add for k in label_names}
        roi_vals[node] = top_feats[node]
        roi_threshold = top_feats[node] + (add / 2)

    else:
        roi_vals = {node: top_feats[node]}
        roi_threshold = None

    # Cast to series
    roi_vals = pd.Series(roi_vals)
    
    # TODO this should be based off plotting vol or surf
    y_adj = 0
    if ref is None:
        y_adj = .05

    # Add reference plot
    ref_ax = fig.add_subplot(gs[i*2, 0])
    ref_ax.text(0, 1, f'{i}). {node}', size=24)
    ref_ax.text(.5, y_adj, f'Weighted Degree: {top_feats[node]:.4f}',
                verticalalignment='center',
                horizontalalignment='center',
                fontstyle='italic')

    # W/ brain
    _plot_brain(roi_vals, fig, ref_ax, ref, label_names,
                show_abs=show_abs, is_top=False, colorbar=False,
                threshold=roi_threshold, **plot_kwargs)
    ref_ax.patch.set_alpha(0)

def _add_row(fig, gs, i, top_feats, ref,
             label_names, show_abs, G, G_norm,
             sub_sz, graph_x_scale=3, graph_y_scale=1, **plot_kwargs):
    
    # Get info needed for passed i
    node = top_feats.index[i-1]

    # Plot reference brain
    _plot_ref(ref=ref, node=node, top_feats=top_feats,
              gs=gs, i=i,
              label_names=label_names, fig=fig,
              show_abs=show_abs, **plot_kwargs)

    # Find x closest nodes to plot sub-graph for sub graph
    path_lens = nx.shortest_path_length(G_norm, source=node, weight='weight')
    sub_nodes = pd.Series(path_lens).sort_values()[:sub_sz].index
    
    # Make plot w/ base G, not G norm
    sub_G = nx.subgraph(G, sub_nodes)
    
    # Add so it goes up into the line ax
    net_ax = fig.add_subplot(gs[(i*2)-1:(i*2)+1, 1])
    net_ax.set_title(f'Subgraph of closest {sub_sz} nodes',
                     size=12, y=-.075, fontstyle='italic')
    add_subgraph_plot(sub_G, G, net_ax, edge_scale=5,
                      layout='circular', x_scale=graph_x_scale,
                      y_scale=graph_y_scale)
    net_ax.patch.set_alpha(0)

    # Get series of all edges from current node - use G
    edges = G.edges(node, data=True)
    edges_vals = pd.Series({e[1] : e[2]['weight'] for e in edges})

    # Add edge's projection plot
    edge_ax = fig.add_subplot(gs[i*2, 2])
    edge_ax.set_title(f'Projection of all edges from {node}', size=14)
    
    _plot_brain(edges_vals, fig, edge_ax, ref, label_names,
                show_abs=show_abs, is_top=False, colorbar=True, **plot_kwargs)
    edge_ax.patch.set_alpha(0)

def _add_bars(degrees, fig, gs, feat_lim=20):

    # Get degree df
    top_feats = degrees.abs().sort_values()[::-1][:feat_lim].index
    df = degrees[top_feats].to_frame().reset_index()
    df = df.rename({'index': 'Node', 0: 'Weighted Degree'}, axis=1)

    # Use full degrees as reference for getting colors
    weights = np.array(df['Weighted Degree'])
    ref_weights = np.array(degrees)
    colors = get_colors(weights, ref_weights)
    palette = {w: c for w, c in zip(df['Node'], colors)}

    # Plot
    end_ax = fig.add_subplot(gs[-1, :])
    sns.barplot(y='Weighted Degree', x='Node', orient='v', data=df, ax=end_ax, palette=palette)
    end_ax.tick_params(axis='x', labelrotation=87)
    sns.despine(ax=end_ax, top=True, left=True, bottom=True, right=True)
    end_ax.patch.set_alpha(0)

    # Add line at 0
    if np.min(weights) < 0:
        end_ax.axhline(0, color='k', lw=.2)

    # Add grid lines
    end_ax.grid(axis='y', linewidth=1, alpha=.33)

def _get_norm_betas(betas):

    # Abs, and prepare for input
    abs_flat_betas = np.abs(betas).flatten().reshape((-1, 1))

    # Scale to 0-1
    scaled_betas = MinMaxScaler().fit_transform(abs_flat_betas).reshape(betas.shape)

    # Returned normed, so lower weight is higher
    return 1 - scaled_betas

def show_network_fis(fis, label_names, ref=None,
                     top_n=3, sub_sz=5, show_abs=True,
                     feat_lim=20,
                     graph_x_scale=2.75, graph_y_scale=1,
                     text_wrap=35, width=26,
                     v_scale=4.25, **plot_kwargs):

    # Optionally apply text-wrap to long labels
    if label_names is not None and text_wrap is not None:
        label_names = np.array(['-\n'.join(wrap(x, text_wrap)) for x in label_names])

    # Process beta's - assume that we need mean here... (for now)
    betas = to_sym(np.array(fis.mean()))

    if show_abs:
        betas = np.abs(betas)

    # Gen networks + degrees
    G = _beta_to_G(betas, label_names, n_std=None)
    G_norm = _beta_to_G(_get_norm_betas(betas), label_names, n_std=None)

    # Base top feats off top abs degrees, if already abs, then nothing changes
    degrees = pd.Series(dict(nx.degree(G, weight='weight'))).sort_values()
    top_feats = degrees.abs().sort_values()[::-1][:top_n].index

    # Init figure and main Grid Spec
    fig = plt.figure(figsize=(width, v_scale * (top_n+2)))
    
    # The base sz of the dotted line in the plot
    # where size is relative to 1.
    line_sz = .1

    # Init with top figure
    height_ratios = [1]
    for feat in top_feats:

        # Make adjustments for text wrap
        line_adj_sz = line_sz + (feat.count('\n') * line_sz * 2)
        height_ratios += [line_adj_sz, 1]

    # Add last line + bar chart
    height_ratios += [line_sz, 2]

    gs = GridSpec(nrows=(top_n*2)+3, ncols=3, figure=fig,
                  width_ratios=[1.25,  1.25, 1.75],
                  height_ratios=height_ratios,
                  hspace=.15, wspace=-.05)

    # Add top row - main degree plot
    _add_top_figure(degrees, gs=gs, fig=fig,
                    ref=ref, label_names=label_names,
                    show_abs=show_abs, **plot_kwargs)

    # Fill in rest of rows
    for i in range(1, top_n+1):

        _add_row(fig, gs, i, degrees.loc[top_feats], ref,
                label_names, show_abs,
                G=G, G_norm=G_norm, 
                sub_sz=sub_sz,
                graph_x_scale=graph_x_scale,
                graph_y_scale=graph_y_scale,
                **plot_kwargs)
        
        _line(fig, gs, (i*2)+1)

    # Add seaborn bar chart importances
    _add_bars(degrees, fig, gs, feat_lim=feat_lim)







"""
Base class for decomposition estimators, utilities for masking and dimension
reduction of group data
"""
import numpy as np
from scipy import linalg
from math import ceil
from sklearn.base import BaseEstimator, TransformerMixin
from joblib import Memory, Parallel, delayed
from sklearn.linear_model import LinearRegression
from sklearn.utils import check_random_state
from sklearn.utils.extmath import randomized_svd, svd_flip

from ..transform.rois import SurfMaps

import warnings
with warnings.catch_warnings():
    warnings.simplefilter(action='ignore', category=FutureWarning)
    from nilearn.signal import _row_sum_of_squares


def fast_svd(X, n_components, random_state=None):
    """ Automatically switch between randomized and lapack SVD (heuristic
        of scikit-learn).
    Parameters
    ----------
    X : array, shape (n_samples, n_features)
        The data to decompose
    n_components : integer
        The order of the dimensionality of the truncated SVD
    random_state : int or RandomState, optional
        Pseudo number generator state used for random sampling.
    Returns
    -------
    U : array, shape (n_samples, n_components)
        The first matrix of the truncated svd
    S : array, shape (n_components)
        The second matrix of the truncated svd
    V : array, shape (n_components, n_features)
        The last matrix of the truncated svd
    """
    random_state = check_random_state(random_state)

    # Small problem, just call full PCA
    if max(X.shape) <= 500:
        svd_solver = 'full'
    elif n_components >= 1 and n_components < .8 * min(X.shape):
        svd_solver = 'randomized'
    # This is also the case of n_components in (0,1)
    else:
        svd_solver = 'full'

    # Call different fits for either full or truncated SVD
    if svd_solver == 'full':
        U, S, V = linalg.svd(X, full_matrices=False)
        # flip eigenvectors' sign to enforce deterministic output
        U, V = svd_flip(U, V)
        # The "copy" are there to free the reference on the non reduced
        # data, and hence clear memory early
        U = U[:, :n_components].copy()
        S = S[:n_components]
        V = V[:n_components].copy()
    else:
        n_iter = 'auto'

        U, S, V = randomized_svd(X, n_components=n_components,
                                 n_iter=n_iter,
                                 flip_sign=True,
                                 random_state=random_state)
    return U, S, V


def mask_and_reduce(imgs,
                    reduction_ratio='auto',
                    n_components=None,
                    random_state=None,
                    n_jobs=1):

    if reduction_ratio == 'auto':
        if n_components is None:
            # Reduction ratio is 1 if
            # neither n_components nor ratio is provided
            reduction_ratio = 1
    else:
        if reduction_ratio is None:
            reduction_ratio = 1
        else:
            reduction_ratio = float(reduction_ratio)
        if not 0 <= reduction_ratio <= 1:
            raise ValueError('Reduction ratio should be between 0. and 1.,'
                             'got %.2f' % reduction_ratio)

    if reduction_ratio == 'auto':
        n_samples = n_components
        reduction_ratio = None
    else:
        # We'll let _mask_and_reduce_single decide on the number of
        # samples based on the reduction_ratio
        n_samples = None

    data_list = Parallel(n_jobs=n_jobs)(
        delayed(_mask_and_reduce_single)(
            img.T,
            reduction_ratio=reduction_ratio,
            n_samples=n_samples,
            random_state=random_state
        ) for img in imgs)
    
    
    return np.vstack(data_list)

def _mask_and_reduce_single(img,
                            reduction_ratio=None,
                            n_samples=None,
                            random_state=None):
    """Utility function for multiprocessing from MaskReducer"""
    random_state = check_random_state(random_state)

    data_n_samples = img.shape[0]
    if reduction_ratio is None:
        assert n_samples is not None
        n_samples = min(n_samples, data_n_samples)
    else:
        n_samples = int(ceil(data_n_samples * reduction_ratio))

    U, S, V = fast_svd(img, n_samples, random_state=random_state)
    U = U.T.copy()
    U = U * S[:, np.newaxis]
    return U


class BaseDecomposition(BaseEstimator, TransformerMixin):
    
    def __init__(self, n_components=20,
                 random_state=None,
                 n_jobs=1,
                 verbose=0):

        self.n_components = n_components
        self.random_state = random_state
        self.n_jobs = n_jobs
        self.verbose = verbose

    def fit(self, imgs, y=None):
        
        data = mask_and_reduce(
                    imgs,
                    reduction_ratio='auto',
                    n_components=self.n_components,
                    random_state=self.random_state,
                    n_jobs=self.n_jobs)

        self._raw_fit(data)

        # Create and fit NiftiMapsMasker for transform
        # and inverse_transform
        self.surf_maps_ = SurfMaps(maps=self.components_.T,
                                   strategy='ls', vectorize=False)

        return self

    def _check_components_(self):
        if not hasattr(self, 'components_'):
            raise ValueError("Object has no components_ attribute. "
                             "This is probably because fit has not "
                             "been called.")

    def transform(self, imgs):
        self._check_components_()
        return np.array([self.surf_maps_.transform(img) for img in imgs])


def explained_variance(X, components, per_component=True):
    """Score function based on explained variance
        Parameters
        ----------
        X : ndarray
            Holds single subject data to be tested against components.
        components : array-like
            Represents the components estimated by the decomposition algorithm.
        per_component : boolean, optional
            Specify whether the explained variance ratio is desired for each
            map or for the global set of components_.
            Default=True.
        Returns
        -------
        score : ndarray
            Holds the score for each subjects. score is two dimensional if
            per_component = True.
        """
    full_var = np.var(X)
    n_components = components.shape[0]
    S = np.sqrt(np.sum(components ** 2, axis=1))
    S[S == 0] = 1
    components = components / S[:, np.newaxis]
    projected_data = components.dot(X.T)
    if per_component:
        res_var = np.zeros(n_components)
        for i in range(n_components):
            res = X - np.outer(projected_data[i],
                               components[i])
            res_var[i] = np.var(res)
            # Free some memory
            del res
        return np.maximum(0., 1. - res_var / full_var)
    else:
        lr = LinearRegression(fit_intercept=True)
        lr.fit(components.T, X.T)
        res = X - lr.coef_.dot(components)
        res_var = _row_sum_of_squares(res).sum()
        return np.maximum(0., 1. - res_var / _row_sum_of_squares(X).sum())


import os

from .. import data_dr

def get_space_options():
    
    spaces = []
    
    for file in os.listdir(data_dr):
        if file.startswith('_'):
            continue
        if file in ['resample_fsaverage', 'mappings', 'parcels', 'index_maps']:
            continue
        
        spaces.append(file)
    
    return spaces


import numpy as np
import pandas as pd
from ..permute_blocks import block_permutation, get_auto_vg

def test_series_cases_base():

    x = pd.Series([1, 2, 3], index=['s1', 's2', 's3'])
    blocks = pd.Series([1, 2, 2], index=['s1', 's2', 's3'])

    # Only s2 and s3 can swap, just need to check s1 then
    for i in range(10):
        p_x = block_permutation(x, blocks)
        assert p_x.loc['s1'] == 1

def test_series_cases_mismatch_order():

    x = pd.Series([1, 2, 3], index=['s1', 's2', 's3'])
    blocks = pd.Series([1, 1, 2], index=['s2', 's3', 's1'])

    # Only s2 and s3 can swap, just need to check s1 then
    for _ in range(10):
        p_x = block_permutation(x, blocks)
        assert p_x.loc['s1'] == 1

def test_df_case():

    x = pd.DataFrame([1, 2, 3], index=['s1', 's2', 's3'])
    x[1] = [1, 2, 3]
    x[2] = [1, 2, 3]

    for _ in range(5):
        p_x = block_permutation(x, x[1])
        assert (p_x == x).sum().sum() ==  9

def test_basic_example():

    blocks = np.array([[-1, 1, -1, 1],
                       [-1, 1, -1, 1],
                       [-1, 1, -2, 1],
                       [-1, 1, -2, 1],
                       [-1, 2, -3, 2],
                       [-1, 2, -3, 2],
                       [-1, 2, -3, 1],
                       [-1, 2, -4, 2],
                       [-1, 2, -4, 1],
                       [-1, 2, -4, 2],
                       [-1, 3, -5, 3],
                       [-1, 3, -6, 3],
                       [-1, 3, -7, 3]])

    indx = np.arange(blocks.shape[0])
    
    # Permute randomly 500 times, testing
    # different constraints.

    spot0 = []
    for random_state in range(500):

        new = block_permutation(x=indx, blocks=blocks, random_state=random_state)
        spot0.append(new[0])
        
        # Make sure main blocks didn't swap
        assert 0 in new[:4]
        assert 1 in new[:4]
        assert 2 in new[:4]
        assert 3 in new[:4]
        assert 4 in new[4:10]
        assert 5 in new[4:10]
        assert 6 in new[4:10]
        assert 7 in new[4:10]
        assert 8 in new[4:10]
        assert 9 in new[4:10]
        assert 10 in new[10:]
        assert 11 in new[10:]
        assert 12 in new[10:]

        # These can only swap with each other
        assert new[4] in [4, 5, 7, 9]
        assert new[5] in [4, 5, 7, 9]
        assert new[7] in [4, 5, 7, 9]
        assert new[9] in [4, 5, 7, 9]
        assert new[6] in [6, 8]
        assert new[8] in [6, 8]
    
    # First index should get all of these
    assert 0 in spot0
    assert 1 in spot0
    assert 2 in spot0
    assert 3 in spot0

def assert_vgs_equal(vg1, vg2):
    '''Comparison between two vgs,
    make sure they are functionally the same'''
    
    # Same number of unique vgs
    assert len(np.unique(vg1)) == len(np.unique(vg2))
    
    # Go through each set of unique groups from vg1
    for u in np.unique(vg1):
        u_inds = np.where(vg1==u)
        
        # Make sure at the same spot in the second
        # vg, it has all of the same number
        assert len(np.unique(vg2[u_inds]))

def test_auto_vg1():

    ex_block = np.array([[-1, 1, 1],
                        [-1, 1, 2],
                        [-1, 2, 1],
                        [-1, 2, 2],
                        [-1, 3, 1],
                        [-1, 3, 2]])

    true_vg = np.array([1, 1, 2, 2, 3, 3])
    auto_vg = get_auto_vg(ex_block)
    assert_vgs_equal(true_vg, auto_vg)

def test_auto_vg2():
    '''More complex example from permute test'''

    blocks = np.array([[-1, 1, -1, 1],
                       [-1, 1, -1, 1],
                       [-1, 1, -2, 1],
                       [-1, 1, -2, 1],
                       [-1, 2, -3, 2],
                       [-1, 2, -3, 2],
                       [-1, 2, -3, 1],
                       [-1, 2, -4, 2],
                       [-1, 2, -4, 1],
                       [-1, 2, -4, 2],
                       [-1, 3, -5, 3],
                       [-1, 3, -6, 3],
                       [-1, 3, -7, 3]])


    true_vg = np.array([1, 1, 1, 1, 2, 2, 3, 2, 3, 2, 4, 4, 4])
    auto_vg = get_auto_vg(blocks)
    assert_vgs_equal(true_vg, auto_vg)


def test_auto_vg3():

    ex_block = np.array([[1, 1, 1],
                         [1, 1, 1],
                         [1, 1, 2],
                         [1, 1, 2],
                         [1, 2, 1],
                         [1, 2, 1],
                         [1, 2, 2],
                         [1, 2, 2]])

    true_vg = np.array([1, 2, 1, 2, 1, 2, 1, 2])
    auto_vg = get_auto_vg(ex_block)
    assert_vgs_equal(true_vg, auto_vg)

def test_auto_vg4():

    ex_block = np.array([[1, 1],
                         [1, 1],
                         [1, 1],
                         [1, 2],
                         [1, 2],
                         [1, 2]])

    true_vg = np.array([1, 2, 3, 1, 2, 3])
    auto_vg = get_auto_vg(ex_block)
    assert_vgs_equal(true_vg, auto_vg)



def test_auto_vg5():

    ex_block = np.array([[1, 1],
                         [1, 1],
                         [1, 1],
                         [1, 2],
                         [1, 2],
                         [1, 2],
                         [1, 3],
                         [1, 3],
                         [1, 3]])

    true_vg = np.array([1, 2, 3, 1, 2, 3, 1, 2, 3])
    auto_vg = get_auto_vg(ex_block)
    assert_vgs_equal(true_vg, auto_vg)


def test_auto_vg6():
    
    # Default case is to assume the first
    # column is within block shuffling
    ex_block = np.array([[1],
                         [1],
                         [1],
                         [2],
                         [2],
                         [2]])

    true_vg = np.array([1, 1, 1, 2, 2, 2])
    auto_vg = get_auto_vg(ex_block)
    assert_vgs_equal(true_vg, auto_vg)


def test_auto_vg7():

    ex_block = np.array([[-1, 1],
                         [-1, 1],
                         [-1, 1],
                         [-1, 2],
                         [-1, 2],
                         [-1, 2]])

    true_vg = np.array([1, 1, 1, 2, 2, 2])

    auto_vg = get_auto_vg(ex_block)
    assert_vgs_equal(true_vg, auto_vg)

def test_auto_vg7():

    ex_block = np.array([[-1, 2],
                         [-1, 2],
                         [-1, 2],
                         [-1, 1],
                         [-1, 1],
                         [-1, 1]])

    true_vg = np.array([1, 1, 1, 2, 2, 2])

    auto_vg = get_auto_vg(ex_block)
    assert_vgs_equal(true_vg, auto_vg)

def test_auto_vg8():

    ex_block = np.array([[-1, 1],
                         [-1, 2],
                         [-1, 1],
                         [-1, 2],
                         [-1, 1],
                         [-1, 2]])

    true_vg = np.array([1, 2, 1, 2, 1, 2])
    auto_vg = get_auto_vg(ex_block)
    assert_vgs_equal(true_vg, auto_vg)


def test_auto_vg9_1D_input():
    '''Should handle passed as 1D array case'''

    ex_block = np.array([1, 1, 1, 2, 2, 2])
    true_vg = np.array([1, 1, 1, 2, 2, 2])
    auto_vg = get_auto_vg(ex_block)
    assert_vgs_equal(true_vg, auto_vg)

def test_auto_vg_10():

    # Is this right?
    ex_block = np.array([[-1, 1, 1],
                         [-1, 1, 1],
                         [-1, 1, 2],
                         [-1, 1, 2],
                         [-1, 2, 1],
                         [-1, 2, 2]])

    true_vg = np.array([1, 2, 1, 2, 3, 3])
    auto_vg = get_auto_vg(ex_block)
    assert_vgs_equal(true_vg, auto_vg)

def test_auto_vg_11():
    ex_block = np.array([[1, 1],
                         [1, 2],
                         [1, 3],
                         [1, 4],
                         [1, 5],
                         [1, 6]])

    true_vg = np.array([1, 1, 1, 1, 1, 1])
    auto_vg = get_auto_vg(ex_block)
    assert_vgs_equal(true_vg, auto_vg)

def test_auto_vg_12():
    ex_block = np.array([[1, 1],
                         [1, 1],
                         [1, 2],
                         [1, 2],
                         [1, 3],
                         [1, 3]])

    true_vg = np.array([1, 2, 1, 2, 1, 2])
    auto_vg = get_auto_vg(ex_block)
    assert_vgs_equal(true_vg, auto_vg)

def test_auto_vg_13():
    ex_block = np.array([[1, -1, 1],
                         [1, -1, 2],
                         [1, -2, 1],
                         [1, -2, 2],
                         [1, -3, 1],
                         [1, -3, 2]])

    true_vg = np.array([1, 2, 1, 2, 1, 2])
    auto_vg = get_auto_vg(ex_block)
    assert_vgs_equal(true_vg, auto_vg)

def test_auto_vg_14():
    ex_block = np.array([[1, -1, 1],
                         [1, -1, 1],
                         [1, -1, 2],
                         [1, -2, 1],
                         [1, -2, 1],
                         [1, -2, 2],
                         [1, -3, 1],
                         [1, -3, 2]])

    true_vg = np.array([1, 1, 2, 1, 1, 2, 3, 4])
    auto_vg = get_auto_vg(ex_block)
    assert_vgs_equal(true_vg, auto_vg)

def test_auto_vg_15():
    ex_block = np.array([[1, -1, 1],
                         [1, -1, 1],
                         [1, -1, 2],
                         [1, -2, 1],
                         [1, -2, 1],
                         [1, -2, 2],
                         [1, -3, 1],
                         [1, -3, 1]])

    true_vg = np.array([1, 1, 2, 1, 1, 2, 3, 3])
    auto_vg = get_auto_vg(ex_block)
    assert_vgs_equal(true_vg, auto_vg)


import numpy as np
import os
import pandas as pd
from ...loading.funcs import load
from ..funcs import run_rely, _test_split
import shutil

file_dr = os.path.dirname(os.path.realpath(__file__))

def setup_test_data():

    fake_dr = os.path.join(file_dr, 'test_data')
    
    # Only save if doesnt exist
    if not os.path.exists(fake_dr):
        os.makedirs(fake_dr, exist_ok=True)

        for i in range(100):
            test_data = np.random.random(10)
            np.save(os.path.join(fake_dr, str(i) + '.npy'), test_data)
    
    # Gen covars df
    covars_df = pd.DataFrame(np.random.random((100, 2)))

    return fake_dr, covars_df


def test_rely_basic():
    
    # Setup fake data
    fake_dr, covars_df = setup_test_data()

    # Just shouldn't fail
    run_rely(covars_df,
             template_path=os.path.join(fake_dr, 'SUBJECT.npy'),
             max_size=30, verbose=-1)

    shutil.rmtree(os.path.join(file_dr, 'test_data'))


def test_rely_basic_data_df():

    covars_df = pd.DataFrame(np.random.random((100, 2)))
    data_df = pd.DataFrame(np.random.random((100, 20)))

    run_rely(covars_df,
             data_df=data_df,
             max_size=30, verbose=-1)

def test_test_split_groups():

    all_subjects = ['s1', 's2', 's3', 's4', 's5']
    stratify = None
    groups = pd.Series([1, 1, 2, 2, 3], index=all_subjects)

    for split in range(50):

        g1, g2 = _test_split(all_subjects, stratify=stratify,
                             groups=groups, split_random_state=split)

        if 's1' in g1:
            assert 's2' in g1
        elif 's1' in g2:
            assert 's2' in g2

        if 's3' in g1:
            assert 's4' in g1
        elif 's3' in g2:
            assert 's4' in g2


def test_test_split_stratify():

    all_subjects = ['s1', 's2', 's3', 's4', 's5', 's6', 's7', 's8']
    stratify = pd.Series([1, 1, 1, 1, 0, 0, 0, 0], index=all_subjects)
    groups = None

    for split in range(50):

        g1, g2 = _test_split(all_subjects, stratify,
                             groups, split_random_state=split)
        assert np.sum(stratify.loc[g1]) == np.sum(stratify.loc[g2])




import os
import nibabel as nib
from tempfile import gettempdir
import random
import numpy as np
from ..loading import load
import shutil
from nibabel.freesurfer.io import read_geometry
from ..transform.formats import geo_to_gifti, data_to_gifti
from .. import data_dr

resample_dr = os.path.join(data_dr, 'resample_fsaverage')

def _proc_temp_dr(temp_dr):

    if temp_dr is None:

        # Ensure random temp directory
        np.random.seed(random.randint(1, 10000000))
        temp_dr = os.path.join(gettempdir(), str(np.random.random()))
        os.makedirs(temp_dr)

    return temp_dr

def gen_cheap_midthickness(fs_subj, hemi, temp_dr=None):
    '''The “correct” way to generate a midthickness file is something like
    mris_expand -thickness lh.white 0.5 graymid.
    In a surf/ directory, this will create a lh.graymid file.
    The cheap way to generate a midthickness file
    is to average the coordinates from white and pial.

    This method does it the cheap way, then saves as gifti to the temp dr.
    
    '''

    temp_dr = _proc_temp_dr(temp_dr)

    # If str, assume directory if dict, paths
    if isinstance(fs_subj, str):
        fs_white = os.path.join(fs_subj, 'surf', f'{hemi}.white')
        fs_pial = os.path.join(fs_subj, 'surf', f'{hemi}.pial')
    elif isinstance(fs_subj, dict):
        fs_white, fs_pial = fs_subj['fs_white'], fs_subj['fs_pial']
    else:
        raise RuntimeError('Must pass fs_subj as directory loc or dict.')
    
    # Load
    white = read_geometry(fs_white)
    pial = read_geometry(fs_pial)
    
    # Get averaged coords
    mid_thickness_coords = np.average([white.coordinates, pial.coordinates], axis=0)
    
    # Conv to gifti
    mid_thickness_gii = geo_to_gifti(mid_thickness_coords, white.faces)
    
    # Save
    mid_thickness_out_loc = os.path.join(temp_dr, f'{hemi}.midthickness.surf.gii')
    nib.save(mid_thickness_gii, mid_thickness_out_loc)
    
    # Return location
    return mid_thickness_out_loc

def _geo_to_gii(in_loc, out_loc):

    in_geo = read_geometry(in_loc)
    in_geo_gii = geo_to_gifti(in_geo[0], in_geo[1])
    nib.save(in_geo_gii, out_loc)

def _get_curr_new_sphere_locs(fs_subj, hemi, target_res, temp_dr):
    '''Get as gifti - saving to temp dr if needed'''

    # If str, assume directory if dict, paths
    if isinstance(fs_subj, str):
        current_fs_sphere = os.path.join(fs_subj, 'surf', f'{hemi}.sphere.reg')
    elif isinstance(fs_subj, dict):
        current_fs_sphere = fs_subj['current_fs_sphere']
    else:
        raise RuntimeError('Must pass fs_subj as directory loc or dict.')

    # Convert current fs sphere to gifti and save
    current_gii_sphere_loc = os.path.join(temp_dr, f'{hemi}.sphere.reg.surf.gii')
    _geo_to_gii(current_fs_sphere, current_gii_sphere_loc)
    
    # Get location of the sphere to sample to
    new_sphere_loc = os.path.join(resample_dr, f'fs_LR-deformed_to-fsaverage.{hemi[0].upper()}.sphere.{target_res}_fs_LR.surf.gii')

    return current_gii_sphere_loc, new_sphere_loc

def freesurfer_resample_prep(fs_subj, hemi,
                             target_res='32k',
                             midthickness_loc=None,
                             temp_dr=None):
    '''This method for now requires that the hcp workbench be installed.'''
    
    temp_dr = _proc_temp_dr(temp_dr)
    
    # If not passed, generate with cheap method
    if midthickness_loc is None:
        midthickness_loc = gen_cheap_midthickness(fs_subj=fs_subj,
                                                  hemi=hemi,
                                                  temp_dr=temp_dr)
    
    # TODO process if midthickness is not already gifti
    
    # Get cur + new_sphere locs
    current_gii_sphere_loc, new_sphere_loc =\
        _get_curr_new_sphere_locs(fs_subj=fs_subj, hemi=hemi, target_res=target_res, temp_dr=temp_dr)

    # Get location of where to save sampled mid thickness
    midthickness_new_out_loc = os.path.join(temp_dr, f'{hemi}.midthickness.{target_res}_fs_LR.surf.gii')

    # Run command
    os.system(f'wb_command -surface-resample {midthickness_loc} {current_gii_sphere_loc} {new_sphere_loc} BARYCENTRIC {midthickness_new_out_loc}')

    # Return dict of created files
    return {'current_gii_sphere_loc': current_gii_sphere_loc,
            'midthickness_new_out_loc': midthickness_new_out_loc,
            'midthickness_loc': midthickness_loc,
            'new_sphere_loc': new_sphere_loc}

def _save_data_as_gii(data, temp_dr=None, append=''):

    temp_dr = _proc_temp_dr(temp_dr)
    
    # If data is list special case
    if isinstance(data, list):
        data_locs = [_save_data_as_gii(d, temp_dr, append=i) for i, d in enumerate(data)]
        return data_locs
    
    # Make sure loaded
    data = load(data)

    # Convert from array to gifti
    data_gii = data_to_gifti(data)

    # Save in temp dr
    data_loc = os.path.join(temp_dr, f'data{append}.gii')
    nib.save(data_gii, data_loc)

    return data_loc

def resample_fs_native_to_fs_LR(data, fs_subj, hemi,
                                target_res='32k',
                                midthickness_loc=None,
                                area_correction=True,
                                temp_dr=None):
    
    # Init temp dr
    temp_dr = _proc_temp_dr(temp_dr)
    
    # Load data to sample as array
    # then save as gifti
    # If data is explicitly a list, data loc will return a list
    data_loc = _save_data_as_gii(data, temp_dr)

    # Where to save result of re-sample
    metric_out_loc = os.path.join(temp_dr, 'metric_out.func.gii')

    if area_correction:

        # Call freesurfer prep to generate files for resample
        files = freesurfer_resample_prep(fs_subj=fs_subj, hemi=hemi,
                                         target_res=target_res,
                                         midthickness_loc=midthickness_loc,
                                         temp_dr=temp_dr)
    
        # Generate command
        cmd = f'wb_command -metric-resample DATA_LOC {files["current_gii_sphere_loc"]} {files["new_sphere_loc"]} ADAP_BARY_AREA {metric_out_loc} -area-surfs {files["midthickness_loc"]} {files["midthickness_new_out_loc"]}'

    else:

        # Get cur + new_sphere locs
        current_gii_sphere_loc, new_sphere_loc =\
        _get_curr_new_sphere_locs(fs_subj=fs_subj, hemi=hemi, target_res=target_res, temp_dr=temp_dr)

        # Generate command
        cmd = f'wb_command -metric-resample DATA_LOC {current_gii_sphere_loc} {new_sphere_loc} BARYCENTRIC {metric_out_loc}'
    
    # If processing multiple
    if isinstance(data_loc, list):

        resampled_data = []
        for d_loc in data_loc:
        
            # Clear existing in case of errors
            if os.path.exists(metric_out_loc):
                os.remove(metric_out_loc)
            
            # Run re-sampling
            os.system(cmd.replace('DATA_LOC', d_loc))
            
            # Load to resampled data
            resampled_data.append(load(metric_out_loc))
    
    # Single data point
    else:

        # Run re-sampling
        os.system(cmd.replace('DATA_LOC', data_loc))
        
        # Load resamples
        resampled_data = load(metric_out_loc)

    # Clean up temp directory
    shutil.rmtree(temp_dr)
    
    # Return the re-sampled data
    return resampled_data


import os
import numpy as np
from .. import data_dr

def load_32k_fs_LR_concat(parcel_name):
    '''Dedicated loader function for saved parcels as generated in
    parc_scaling project. These parcellations are all in space 32k_fs_LR,
    and are left, right hemisphere concatenated. The parcel will be
    downloaded to parcels directory in the default data dr.

    Parameters
    -----------
    parcel_name : str
        The name of the parcel to load,
        see https://github.com/sahahn/parc_scaling/tree/main/parcels
        for valid options.
    
    Returns
    -------
    parcel : numpy array
        The loaded concat LR parcellation as a numpy array is returned.
    '''

    # Tolerance to passing with or without npy tag
    parcel_name = parcel_name.replace('.npy', '')
    
    # Make sure parcel directory init'ed
    parcel_dr = os.path.join(data_dr, 'parcels', '32k_fs_LR_concat')
    os.makedirs(parcel_dr, exist_ok=True)
    
    # Get location
    parcel_loc = os.path.join(parcel_dr, f'{parcel_name}.npy')
    
    # If doesn't exist, try to download
    if not os.path.exists(parcel_loc):
        print('Downloading parcel...')
        parcel_url = f'https://raw.githubusercontent.com/sahahn/parc_scaling/main/parcels/{parcel_name}.npy'
        os.system(f'wget -L {parcel_url} -P {parcel_dr}')

    # Load parcel
    parcel = np.load(parcel_loc)

    return parcel


import numpy as np
import os
from ..funcs import load
import nibabel as nib
import shutil

file_dr = os.path.dirname(os.path.realpath(__file__))

def test_load_basic():

    os.makedirs(os.path.join(file_dr, 'test_data'), exist_ok=True)

    x = np.ones(10)
    p1 = os.path.join(file_dr, 'test_data', 'ones.npy')
   
    np.save(p1, x)

    assert np.array_equal(load(p1), x)

    p2 = os.path.join(file_dr, 'test_data', 'ones.nii')
    p3 = os.path.join(file_dr, 'test_data', 'ones.nii.gz')
    x_nib = nib.Nifti1Image(x, np.eye(4))
    nib.save(x_nib, p2)
    nib.save(x_nib, p3)

    assert np.array_equal(load(p2), x)
    assert np.array_equal(load(p3), x)

    shutil.rmtree(os.path.join(file_dr, 'test_data'))






import pandas as pd
from sklearn.preprocessing import OrdinalEncoder
from joblib.hashing import hash as joblib_hash
from sklearn.preprocessing import LabelEncoder
from ..misc.print import _get_print
from .cache import _base_cache_load


def _base_load_from_csv(cols, csv_loc, eventname):
    '''The actual loading piece'''

    # Load from csv - using abcd specific NaN values
    data = pd.read_csv(csv_loc,
                       usecols=['src_subject_id', 'eventname'] + cols,
                       na_values=['777', 999, '999', 777])

    # If None, keep all and return
    if eventname is None:
        return data

    # Otherwise, we assume eventname is list
    data = data.loc[data[data['eventname'].isin(eventname)].index]

    # If just one eventname, drop the eventname col
    if len(eventname) == 1:
        data = data.drop('eventname', axis=1)

    return data

def _order_cache_loaded(data, cols):
    '''When loading from hash we make sure to return as ordered
    by cols in the cases where the saved copy was saved with a different
    column order.'''

    # Generate correct ordering
    ordered_cols = ['src_subject_id']

    # Eventname may or may not be there
    if 'eventname' in data:
        ordered_cols += ['eventname']

    # Add rest of cols
    ordered_cols += cols

    # Apply the ordering and return
    return data[ordered_cols]

def _csv_cache_load_func(cache_loc, load_args):

    # Cached data saved as tab seperated, without any index cols
    data = pd.read_csv(cache_loc, sep='\t')

    # Set to correct column order
    data = _order_cache_loaded(data, load_args[0])

    return data

def _csv_cache_save_func(data, cache_loc):

     # Save a copy under the cache_loc as tsv, no index
    data.to_csv(cache_loc, index=False, sep='\t')

def load_from_csv(cols, csv_loc,
                  eventname='baseline_year_1_arm_1',
                  drop_nan=False, encode_cat_as='ordinal',
                  verbose=0, **cache_args):
    '''Special ABCD Study specific helper utility to load specific
    columns from a csv saved version of the DEAP release RDS
    file or similar ABCD specific csv dataset.

    Parameters
    ----------
    cols : str or list-like
        Either a single str with the column name to load,
        or a list / list-like with the names of multiple columns to
        load.

        If any variable passed is wrapped in 'C(variable_name)'
        that variable will be ordinalized (or whatever option is specified
        with the encode_cat_as option) and saved under the base variable name
        (i.e., with C() wrapper removed).

    csv_loc : str
        The str location of the csv saved version of the DEAP release
        RDS file for the ABCD Study - or any other comma seperated dataset
        with an eventname column.

    eventname : str, list-like or None, optional
        The single eventname as a str, or multiple eventnames
        in which to include results by. If passed as None then
        all avaliable data will be kept.

        If a single eventname is specified then the eventname
        column will be dropped, otherwise it will be kept.

        ::

            default = 'baseline_year_1_arm_1'

    drop_nan : bool, optional
        If True, then drop any rows / subjects data with
        missing values in any of the requested columns.

        Note: Any values encoded as ['777', 999, '999', 777]
        will be treated as NaN. This is a special ABCD specific consideration.

        ::

            default = False

    encode_cat_as : {'ordinal', 'one hot', 'dummy'}, optional
        The way in which categorical vars, any wrapped in C(),
        should be categorically encoded.

        - 'ordinal' :
          
          The variables in encoded sequentially in one
          column with the original name, with values 0 to k-1
          where k is the number of unique categorical values.
          This method uses :class:`OrdinalEncoder<sklearn.preprocessing.OrdinalEncoder>`.

        - 'one hot' :
          
          The variables is one hot encoded, adding columns
          for each unique value. This method uses function :func:`pandas.get_dummies`.

        - 'dummy' :
          
          Same as 'one hot', except one of the columns is then dropped.

    cache_args : keyword arguments
        There are a number of optional cache arguments that can be set via
        kwargs, as listed below.

        - **cache_dr** : str or None
        
          The location of where to cache the results of
          this function, for faster loading in the future.

          If None, do not cache. The default if not set is
          'default', which will use caching in a location defined
          by the function name in a folder called
          neurotools_cache in the users homes directory.

        - **cache_max_sz** : str or int
        
          This parameter defines the maximize size of
          the cache directory. The idea is that if saving a new
          cached function call and it exceeds this cache max size,
          previous saved caches (by oldest in terms of used) will be
          deleted, ensuring the cache directory remains under this size.

          Can either pass in terms of bytes directly as a number,
          or in terms of a str w/ byte marker, e.g., '14G' for
          14 gigabytes, or '10 KB' for 10 kilobytes.

          The default if not set is '30G'.

        - **use_base_name** : bool
            
          Optionally when any arguments used in the
          caching can be cached based on either their full file
          path, if use_base_name is False, or just the file name itself,
          so for example /some/path/location vs. just location.
          The default if not set is True, as it assumes that maybe
          another file in another location with the same name is the same.

    Returns
    -------
    df : :class:`pandas.DataFrame`
        Will return a :class:`pandas.DataFrame` as indexed by
        column `src_subject_id` within the original csv.

    '''

    # TODO move default cache location to
    # user's home directory ??? 

    # Get verbose printer
    _print = _get_print(verbose)
    
    # Handle passing input as single str / column
    if isinstance(cols, str):
        cols = [cols]
    if isinstance(eventname, str):
        eventname = [eventname]

    # Make sure proper input for cols and eventname
    # if not list, try to cast to list in case of other list-like
    if not isinstance(cols, list):
        try:
            cols = list(cols)
        except:
            raise RuntimeError(f'Passed cols: {cols} must be passed as str or list-like!')

    # Extra is None case for eventname
    if not isinstance(eventname, list) and eventname is not None:
        try:
            eventname = list(eventname)
        except:
            raise RuntimeError(f'Passed eventname: {eventname} must be passed as str, list-like or None!')

    # Check for any wrapped in C()
    use_cols = cols.copy()
    cat_vars = []
    for i, col in enumerate(use_cols):
        if col.startswith('C(') and col.endswith(')'):
            var_name = col[2:-1]

            # Replace with unwrapped
            use_cols[i] = var_name

            # Keep track
            cat_vars.append(var_name)
    
    # Load with optional caching - only need
    # to cache this main operation.
    data = _base_cache_load(load_args=(use_cols, csv_loc, eventname),
                            load_func=_base_load_from_csv,
                            cache_load_func=_csv_cache_load_func,
                            cache_save_func=_csv_cache_save_func,
                            cache_args=cache_args,
                            _print=_print)

    # Set index as subject id
    data = data.set_index('src_subject_id')

    # Optionally drop NaN's
    if drop_nan:
        original_len = len(data)
        data.dropna(inplace=True)
        _print(f'Dropping {original_len - len(data)} subjects for missing data.', level=1)
    
    # Ordinally encode any C wrapped vars - categorical vars
    _print(f'Categorical vars to be encoded =', cat_vars, level=3)
    if len(cat_vars) > 0:

        _print(f'Encoding {len(cat_vars)} categorical vars with {encode_cat_as} encoding.', level=1)

        if encode_cat_as == 'ordinal':
            data[cat_vars] = OrdinalEncoder().fit_transform(data[cat_vars])
        elif encode_cat_as == 'one hot':
            data = pd.get_dummies(data=data, columns=cat_vars, drop_first=False)
        elif encode_cat_as == 'dummy':
            data = pd.get_dummies(data=data, columns=cat_vars, drop_first=True)
        else:
            raise RuntimeError(f'encode_cat_as {encode_cat_as} not valid option.')    

    return data

def load_family_block_structure(csv_loc, subjects=None,
                                eventname='baseline_year_1_arm_1',
                                add_neg_ones=False, cache_dr='default',
                                cache_max_sz='30G', verbose=0):
    '''This helper utility loads PALM-style exchanability blocks for ABCD study specific data
    according to right now a fixed set of rules:

    - Families of the same type can be shuffled (i.e., same number of members + of same status)
    - Siblings of the same type can be shuffled 
    - Treat DZ as ordinary sibs (i.e., just treat MZ seperately)

    Parameters
    ----------
    csv_loc : str / file path
        The location of the csv saved version of the DEAP release
        RDS file for the ABCD Study. This can also just be any
        other csv as long as it has columns:
        'rel_family_id', 'rel_relationship', 'genetic_zygosity_status_1'

    subjects : None or array-like, optional
        Can optionally specify that the block structure be created on a subset of subjects,
        though if any missing values are present in rel_relationship or rel_family_id
        within this subset, then those will be further dropped.

        If passed as non-null, this should be a valid array-like
        or :class:`pandas.Index` style set of subjects.

        ::

            default = None

    eventname : str, array-like or None, optional
        A single eventname as a str in which to specify data by.

        For now, this method only supports loading data at a single
        time point across subjects.

        ::

            default = 'baseline_year_1_arm_1'

    add_neg_ones : bool, optional
        If True, add a left-most column
        with all negative ones representing
        that swaps should occur within group at
        the outermost level. Note that if using
        a permutation function through neurotools
        that accepts this style of blocks,
        this outer layer is assumed by default,
        so this parameter can be left as False.

        ::

            default = False

    Returns
    --------
    block_structure : :class:`pandas.DataFrame`
        The loaded block structure as indexed by src_subject_id with
        columns: 'neg_ones', 'family_type', 'rel_family_id' and 'rel_relationship'.
        
        Any subjects with missing values in a key column have been dropped from
        this returned structure.
    '''

    # Get verbose printer
    _print = _get_print(verbose)

    # TODO: Add in support for what a block structure with
    # multiple eventnames, i.e., longitudinal data, and what
    #  that would look like.

    # Load in needed columns, w/o NaNs
    rel_cols = ['rel_family_id', 'rel_relationship', 'genetic_zygosity_status_1']
    data = load_from_csv(rel_cols, csv_loc,
                         eventname=eventname,
                         drop_nan=False,
                         cache_dr=cache_dr,
                         cache_max_sz=cache_max_sz,
                         verbose=verbose)
    _print(f'Loaded data needed with shape: {data.shape}', level=1)

    # Set to subset of passed subjects if any,
    if subjects is not None:
        data = data.loc[subjects]
        _print(f'Set to {len(data)} subjects', level=1)

    # Drop any data points if rel_family_id or rel_relationship is missing
    to_drop = data[pd.isnull(data[['rel_relationship',
                                   'rel_family_id']]).any(axis=1)].index
    data = data.drop(to_drop)

    # Treat twins if dy zy as normal siblings, basically if not labelled mono, then normal sib
    treat_as_siblings = data[((data['rel_relationship'] == 'twin') | (data['rel_relationship'] == 'triplet')) &\
                              (data['genetic_zygosity_status_1'] != 'monozygotic')].index
    data.loc[treat_as_siblings, 'rel_relationship'] = 'sibling'

    # Create dict mapping subj to family
    families = {}
    for subj in data.index:
        rel_family_id = data.loc[subj, 'rel_family_id']
        family = data[data['rel_family_id'] == rel_family_id].index
        families[subj] = family

    # Init column
    data['family_type'] = 0
    hashes = {}
    cnt = 0

    # Fill in unique value for every type of family
    for subj in data.index:
        
        # Get unique hash
        h = joblib_hash(data.loc[families[subj], 'rel_relationship'].values)
        
        try:
            u = hashes[h]
        except KeyError:
            hashes[h] = cnt
            u = cnt
            cnt += 1
            
        data.loc[subj, 'family_type'] = u

    # Ordinally encode cols
    data['rel_relationship'] = LabelEncoder().fit_transform(data['rel_relationship'])
    data['rel_family_id'] = LabelEncoder().fit_transform(data['rel_family_id'])

    # Add outer col with -1's if requested
    if add_neg_ones:
        data['neg_ones'] = -1

        return data[['neg_ones', 'family_type', 'rel_family_id', 'rel_relationship']]

    # Default order w/o neg ones
    return data[['family_type', 'rel_family_id', 'rel_relationship']]
